import sys
import json
import shutil
import string
from random import choice

def deunicodeData(data):
  if isinstance(data, list):
    new_data = []
    for sub_data in data:
      new_data.append(deunicodeData(sub_data))
  elif isinstance(data, unicode):
    new_data = data.encode('utf8')
  elif isinstance(data, dict):
    new_data = {}
    for key, value in data.iteritems():
      key = deunicodeData(key)
      value = deunicodeData(value)
      new_data[key] = value
  else:
    new_data = data
  return new_data


##############################################################################
#
# Copyright (c) 2011 Nexedi SA and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly advised to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################
from datetime import datetime,timedelta
import os
import subprocess
import sys
import time
import glob
import SlapOSControler
import json
import time
import shutil
import logging
import string
import random
from ProcessManager import SubprocessError, ProcessManager, CancellationError
from subprocess import CalledProcessError
from Updater import Updater
from erp5.util import taskdistribution

class SlapOSInstance(object):
  """
  Base of an software instance,
  store variables used during software installation
  """
  def __init__(self):
    self.retry_software_count = 0
    self.retry = False

  def edit(self, **kw):
    self.__dict__.update(**kw)
    self._checkData()

  def _checkData(self):
    pass
    
    
class NodeTestSuite(SlapOSInstance):
  """
  
  """
  def __init__(self, reference):
    super(NodeTestSuite, self).__init__()
    self.reference = reference

  def edit(self, **kw):
    super(NodeTestSuite, self).edit(**kw)

  def _checkData(self):
    if getattr(self, "working_directory", None) is not None:
      if not(self.working_directory.endswith(os.path.sep + self.reference)):
        self.working_directory = os.path.join(self.working_directory,
                                             self.reference)
      SlapOSControler.createFolder(self.working_directory)
      self.test_suite_directory = os.path.join(
                                   self.working_directory, "test_suite")
      self.custom_profile_path = os.path.join(self.working_directory,
                                 'software.cfg')
    if getattr(self, "vcs_repository_list", None) is not None:
      for vcs_repository in self.vcs_repository_list:
        buildout_section_id = vcs_repository.get('buildout_section_id', None)
        repository_id = buildout_section_id or \
                        vcs_repository.get('url').split('/')[-1].split('.')[0]
        repository_path = os.path.join(self.working_directory,repository_id)
        vcs_repository['repository_id'] = repository_id
        vcs_repository['repository_path'] = repository_path

  def createSuiteLog(self):
    # /srv/slapgrid/slappartXX/srv/var/log/testnode/az-mlksjfmlk234Sljssdflkj23KSdfslj/suite.log
    alphabets = string.digits + string.letters
    rand_part = ''.join(random.choice(alphabets) for i in xrange(32))
    random_suite_folder_id = '%s-%s' % (self.reference, rand_part)
    suite_log_directory = os.path.join(self.log_directory,
                                       random_suite_folder_id)
    SlapOSControler.createFolders(suite_log_directory)
    self.suite_log_path = os.path.join(suite_log_directory,
                                       'suite.log')
    return self.getSuiteLogPath(), random_suite_folder_id

  def getSuiteLogPath(self):
    return getattr(self,"suite_log_path", None)




import json
import httplib
import urlparse
import time

TIMEOUT = 30

# TODO: News-> look list to get last news... (and not the first of the list)

class SlapOSMasterCommunicator(object):
  """
  Communication with slapos Master using Hateoas.
  
  collection: collection of data (hosting_subscription, instance, software_release...)
  hosting_subscription: result of a request
  instance(s): instance(s) related to an hosting_subscription

  usage: ex:
    # Try to reuse same communicator, because initilization step may takes a lot of time
    # due to listing of all instances (alive or not) related to the specified slapOS account.
    communicator = SlapOSMasterCommunicator()
    
    # Print news related to 'TestScalability_21423104630420' all instances
    instance_link_list =  communicator._getRelatedInstanceLink('TestScalability_21423104630420')
    for instance_link in instance_link_list:
      news = communicator.getNewsFromInstanceLink(instance_link)
      print news['news']
  """
  def __init__(self, certificate_path, key_path, log,
                     url):
    # Create connection
    api_scheme, api_netloc, api_path, api_query, api_fragment = urlparse.urlsplit(url)
    self.log = log
    self.certificate_path = certificate_path
    self.key_path = key_path
    self.url = url
    self.connection = self._getConnection(self.certificate_path, self.key_path, self.url)
    # Get master
    master_link = {'href':api_path,'type':"application/vnd.slapos.org.hal+json; class=slapos.org.master"}
    master = self._curl(master_link)
    self.person_link = master['_links']['http://slapos.org/reg/me']
    # Get person related to specified key/certificate provided
    person = self._curl(self.person_link)
    self.personnal_collection_link = person['_links']['http://slapos.org/reg/hosting_subscription']
    # Get collection (of hosting subscriptions)
    collection = self._curl(self.personnal_collection_link)
    # XXX: This part may be extremly long (because here no hosting subscriptions
    # has been visited)
    self.hosting_subcriptions_dict = {}
    self.visited_hosting_subcriptions_link_list = []
    self.log("SlapOSMasterCommunicator will read all hosting subscriptions entries, "
             "it may take several time...")
    self._update_hosting_subscription_informations()
    
  def _getConnection(self,certificate_path, key_path, url):
    api_scheme, api_netloc, api_path, api_query, api_fragment = urlparse.urlsplit(url)
    #self.log("HTTPS Connection with: %s, cert=%s, key=%s" %(api_netloc,key_path,certificate_path))
    return httplib.HTTPSConnection(api_netloc, key_file=key_path, cert_file=certificate_path, timeout=TIMEOUT)

  def _curl(self, link):
    """
    'link' must look like : {'href':url,'type':content_type}
    """
    # Set timeout
    import socket
    socket.setdefaulttimeout(1.0*TIMEOUT)
    
    api_scheme, api_netloc, api_path, api_query, api_fragment = urlparse.urlsplit(link['href'])
    max_retry = 10
    # Try to use existing conection
    try:
      self.connection.request(method='GET', url=api_path, headers={'Accept': link['type']}, body="")
      response = self.connection.getresponse()
      return json.loads(response.read())
    # Create and use new connection
    except:
      retry = 0
      # (re)Try several time to use new connection
      while retry < max_retry:
        try:
          self.connection = self._getConnection(self.certificate_path, self.key_path, self.url)
          self.connection.request(method='GET', url=api_path, headers={'Accept': link['type']}, body="")
          response = self.connection.getresponse()
          return json.loads(response.read())
        except:
          self.log("SlapOSMasterCommunicator: Connection failed..")
          retry += 1
          time.sleep(10)
    self.log("SlapOSMasterCommunicator: All connection attempts failed after %d try.." %max_retry)
    raise ValueError("SlapOSMasterCommunicator: Impossible to use connection")
        
  def _update_hosting_subscription_informations(self):
    """
    Add all not already visited hosting_subcription
    # Visit all hosting subscriptions and fill a dict containing all
    # new hosting subscriptions. ( like: {hs1_title:hs1_link, hs2_title:hs2_link, ..} )
    # and a list of visited hosting_subsciption ( like: [hs1_link, hs2_link, ..] )
    """
    collection = self._curl(self.personnal_collection_link)
    # For each hosting_subcription present in the collection
    for hosting_subscription_link in collection['_links']['item']:
      if hosting_subscription_link not in self.visited_hosting_subcriptions_link_list:
        hosting_subscription = self._curl(hosting_subscription_link)
        self.hosting_subcriptions_dict.update({hosting_subscription['title']:hosting_subscription_link})
        self.visited_hosting_subcriptions_link_list.append(hosting_subscription_link)
  
  def _getRelatedInstanceLink(self, hosting_subscription_title):
    """
    Return a list of all related instance_url from an hosting_subscription_title
    """
    # Update informations
    self._update_hosting_subscription_informations()
    # Get specified hosting_subscription
    hosting_subscription_link = self.hosting_subcriptions_dict[hosting_subscription_title]
    hosting_subscription = self._curl(hosting_subscription_link)
    assert(hosting_subscription_title == hosting_subscription['title'])
    # Get instance collection related to this hosting_subscription
    instance_collection_link = hosting_subscription['_links']['http://slapos.org/reg/instance']
    instance_collection = self._curl(instance_collection_link)
    related_instance_link_list = []
    # For each instance present in the collection
    for instance in instance_collection['_links']['item']:
      related_instance_link_list.append(instance)
    return related_instance_link_list

  def getNewsFromInstanceLink(self, instance_link):
      instance = self._curl(instance_link)
      news_link = instance['_links']['http://slapos.org/reg/news']
      return self._curl(news_link)

  def isHostingSubsciptionStatusEqualTo(self, hosting_subscription_title, excepted_news_text):
    """
    Return True if all related instance state are equal to status,
    or False if not or if there is are no related instances.
    """
    related_instance_link_list = _getRelatedInstanceLink(hosting_subscription_title)
    # For each instance present in the collection
    for instance_link in related_instance_link_list:
      news = self.getNewsFromInstanceLink(instance_link)
      if excepted_news_text != news['news'][0]['text']:
        return False
    return len(related_instance_link_list) > 0

  def isInstanceReady(self, instance_link, status):
    """
    Return True if instance status and instance news text ~looks corresponding.
    ( use the matching of 'correctly' and 'Instance' and status )
    """
    # XXX: SlapOS Master doesn't store any "news" about slave instances. Assume true.
    if self._curl(instance_link)['slave']:
      return True
    text = self.getNewsFromInstanceLink(instance_link)['news'][0]['text']
    return ('Instance' in text) and ('correctly' in text) and (status in text)

  # check if provided 'status' = status
  def isHostingSubscriptionReady(self, hosting_subscription_title, status):
    """
    Return True if all instance status and instance news text ~looks corresponding.
    ( use the matching of 'correctly' and 'Instance' and status ).
    """
    instance_link_list = self._getRelatedInstanceLink(hosting_subscription_title)
    for instance_link in instance_link_list:
      if not self.isInstanceReady(instance_link, status):
        return False
    return len(instance_link_list) > 0
    
  def isRegisteredHostingSubscription(self, hosting_subscription_title):
    """
    Return True if the specified hosting_subscription is present on SlapOSMaster
    """
    self._update_hosting_subscription_informations()
    if self.hosting_subcriptions_dict.get(hosting_subscription_title):
      return True
    return False

  def getHostingSubscriptionDict(self):
    """
    Return the dict of hosting subcription.
    """
    return self.hosting_subcriptions_dict

  def getHostingSubscriptionInformationDict(self, title):
    """
    Return a dict with informations about Hosting subscription
    """
    related_instance_link_list = self._getRelatedInstanceLink(title)
    related_instance_link = None
    # Get root instance
    for link in related_instance_link_list:
      instance = self._curl(link)
      if title == instance['title']:
        related_instance_link = link
        break
    # Return information dict
    if related_instance_link:
      related_instance = self._curl(related_instance_link)
      return {
        'title': related_instance['title'],
        'status': related_instance['status'],
        'software_url': related_instance['_links']['http://slapos.org/reg/release'],
        'software_type': related_instance['software_type'],
        'computer_guid': related_instance['sla']['computer_guid']
      }
    else:
      return None
  


#!/usr/bin/python
##############################################################################
#
# Copyright (c) 2009 Nexedi SA and Contributors. All Rights Reserved.
#                    Vincent Pelletier <vincent@nexedi.com>
#                    Sebastien Robin <seb@nexedi.com>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################
from datetime import date
from os import path
import rpy2.robjects as robjects
import os
from optparse import OptionParser
r = robjects.r

usage = """
  Usage:
    %prog [OPTION] file1.csv [file2.csv [...]]
  Result:
    Generates, in current directory, a graph per csv column in out-type format.
    Their name is composed of:
    - csv file basename (without extension)
    - csv column title
    - the ratio of present points (100 to 000). The higher the number, the
      more the plot will be complete (less holes, longer timespan coverage).
    - out-type extension

  CSV files must have been generated by parse_timing_log.py tool.
"""

class CSVFile(object):
  def __init__(self, file_name, field_delim=','):
    file = open(file_name, 'r')
    self.column_dict = column_dict = {}
    self.column_list = column_list = []
    self.ratio_dict = ratio_dict = {}
    line_num = 0
    self.value_max = value_max = {}
    next_ord = 0
    for x, title in enumerate(file.readline().split(field_delim)):
      title = title.strip()
      title = title.strip('"')
      if title in column_dict:
        title = next_ord
        while title in column_dict:
          title += 1
        next_ord = title + 1
        title = str(title)
      column_dict[title] = []
      column_list.append(title)
    for line in file.readlines():
      line_num += 1
      for x, cell in enumerate(line.split(field_delim)):
        cell = cell.strip()
        key = column_list[x]
        if x != 0:
          cell = computeExpr(cell)
          if cell is not None:
            ratio = ratio_dict.get(key, 0)
            ratio_dict[key] = ratio + 1
            if cell > value_max.get(key, 0):
              value_max[key] = cell
        column_dict[key].append(cell)
    line_num = float(line_num) / 100
    for key in ratio_dict:
      ratio_dict[key] /= line_num

  def getColumn(self, column_id):
    return self.column_dict[self.column_list[column_id]]

  def iterColumns(self, start=0, stop=None):
    if stop is None:
      column_list = self.column_list[start:]
    else:
      column_list = self.column_list[start:stop]
    return ((x, self.column_dict[x], self.value_max.get(x, 0), self.ratio_dict.get(x, 0)) for x in column_list)

def computeExpr(expr):
  # only supports '=x/y'
  if expr:
    assert expr[0] == '='
    num, denom = expr[1:].split('/')
    result = float(int(num)) / int(denom)
  else:
    result = None
  return result

def main():
  parser = OptionParser(usage)
  parser.add_option("--with-regression", action="store_true",
    dest="regression_enabled", help="enable B-spline regression")
  parser.add_option("--ignored-quantity", type="int", dest="ignored_quantity",
    help="ignore IGNORED_QUANTITY higher values that might make a graph totally unusable")
  parser.add_option("--out-type", type="string", default="png",
    help="can be %default (default) or svg")
  parser.add_option("--minimal-non-empty-values-ratio", type="float",
    dest="minimal_non_empty_ratio", default=None,
    help="graph with ratio of non empty values with lesser than value, then graph is ignored")
  (options, file_name_list) = parser.parse_args()

  current_dir = os.getcwd()
  for file_name in file_name_list:
    print 'Loading %s...' % (file_name, )
    file = CSVFile(file_name)

    date_string_list = file.getColumn(0)
    date_list = []
    x_label_value_list = []
    # plotting functionnalities does not select smartly
    # a good number of x values to display, so we will display 20 dates
    # in order to have good enough dates on the x axis.
    # x_label_value_list will be like [1, 5, 10...]
    # date_list will be like ['2009/07/01', '2009/07/05', '2009/07/10', ...]
    factor = 1
    if len(date_string_list) > 20:
      factor = int(len(date_string_list) / 20)
    i = 0
    for date_string in date_string_list:
      if i % factor == 0:
        x_label_value_list.append(i)
        date_split = date_string.replace('"','').split('/')
        date_split.reverse()
        new_date = '/'.join(date_split)
        date_list.append(new_date)
      i += 1
    max_x = len(date_string_list)
    # knots are used for B-spline regression
    # We need to add three additional knots at the begin and end in
    # order to have the right basis
    knot_list  = [x_label_value_list[0]] * 3 + x_label_value_list \
        + [max_x] * 4
    r_x_label_value_list = robjects.FloatVector(x_label_value_list)
    robjects.globalenv["x_label_value_list"] = r_x_label_value_list
    robjects.globalenv["knot_list"] = knot_list
    r("x_label <- c(%s)" % ','.join(['"%s"' % x for x in date_list]))
    # import the splines library in R
    if options.regression_enabled:
      r("library(splines)")
    # now parse all columns and store a out-type file
    for title, column, value_max, ratio in file.iterColumns(start=1):
      out_file, out_ext = path.splitext(path.basename(file_name))
      if out_ext != '.csv':
        out_file = '.'.join((out_file, out_ext))
      out_file_name = '%s_%s_%03i.%s' % (out_file, title.replace('%',''),
          ratio, options.out_type)
      i = 0
      x_data = []
      y_data = []
      # First parse the list to retrieve values that we might want to remove
      ignored_value_set = set([])
      max_y_data = []
      if options.ignored_quantity not in (None, 0):
        for value in column:
          if value is not None:
            max_y_data.append(value)
        max_y_data.sort()
        ignored_value_set = set(max_y_data[-options.ignored_quantity:])
      # build list with all data that we want to display
      for value in column:
        if value is not None and not (value in ignored_value_set):
          x_data.append(i)
          y_data.append(value)
        i += 1
      if len(x_data) == 0:
        print 'Nothing to plot for %s...' % (out_file_name, )
        continue
      if options.minimal_non_empty_ratio is not None:
        column_len = len(column)
        if column_len:
          if float(len(x_data))/column_len < options.minimal_non_empty_ratio:
            print 'Not enough values to plot for %s...' % (out_file_name, )
            continue
      r_y_data = robjects.FloatVector(y_data)
      r_x_data = robjects.FloatVector(x_data)
      robjects.globalenv["y_data"] = r_y_data
      robjects.globalenv["x_data"] = r_x_data
      display_column_regression = options.regression_enabled
      # if there is no more than one unique point, regression is useless
      if len(set([x for x in r_y_data])) <= 1:
        display_column_regression = 0
      regression_string = ''
      # Calculate a B-spline regression in order to give clear overview
      # about the direction of chaotics values.
      if display_column_regression:
        r("bx <- splineDesign(knot_list, x_data)")
        r("fitted_model <- lm(y_data ~ bx)")
        regression_string = ', fitted_model$fit'
      # Define the place where to store the graphe and format of the image
      r("""%s(file='%s/%s', width=800, height=600)""" % (options.out_type,
        current_dir, out_file_name))
      # Increase the size for the place of the bottom axis labels (x)
      r("""par(mar=c(9, 4, 4, 2) + 0.1)""")
      # Plot the graph itself
      r("""matplot(x_data, cbind(y_data %s), type='ll',
                lty=1, main='%s (average display time per day)',
                xlab='', ylab='time (s)', xaxt='n')""" % (
                  regression_string, title))
      r("""axis(1, at=x_label_value_list, lab=x_label, las=2)""")
      # stop changing the out-type file
      r("""dev.off()""")

      print 'Saving %s...' % (out_file_name, )

if __name__ == '__main__':
  main()




#!/usr/bin/env python2.4

# Based on a transaction analyzer by Matt Kromer.

import sys
import os
import getopt
import anydbm as dbm
import tempfile
import shutil
from ZODB.FileStorage import FileStorage
from ZODB.utils import get_pickle_metadata

class Report:
    def __init__(self, use_dbm=False):
        self.use_dbm = use_dbm
        if use_dbm:
            self.temp_dir = tempfile.mkdtemp()
            self.OIDMAP = dbm.open(os.path.join(self.temp_dir, 'oidmap.db'),
                                   'nf')
            self.USEDMAP = dbm.open(os.path.join(self.temp_dir, 'usedmap.db'),
                                    'nf')
        else:
            self.OIDMAP = {}
            self.USEDMAP = {}
        self.TYPEMAP = {}
        self.TYPESIZE = {}
        self.TIDS = 0
        self.OIDS = 0
        self.DBYTES = 0
        self.COIDS = 0
        self.CBYTES = 0
        self.FOIDS = 0
        self.FBYTES = 0
        self.COIDSMAP = {}
        self.CBYTESMAP = {}
        self.FOIDSMAP = {}
        self.FBYTESMAP = {}

def shorten(s, n):
    l = len(s)
    if l <= n:
        return s
    while len(s) + 3 > n: # account for ...
        i = s.find(".")
        if i == -1:
            # In the worst case, just return the rightmost n bytes
            return s[-n:]
        else:
            s = s[i + 1:]
            l = len(s)
    return "..." + s

def report(rep, csv=False):
    if not csv:
        print "Processed %d records in %d transactions" % (rep.OIDS, rep.TIDS)
        print "Average record size is %7.2f bytes" % (rep.DBYTES * 1.0 / rep.OIDS)
        print ("Average transaction size is %7.2f bytes" %
               (rep.DBYTES * 1.0 / rep.TIDS))

        print "Types used:"
    if csv:
        fmt = "%s,%s,%s,%s,%s,%s,%s,%s,%s"
        fmtp = "%s,%d,%d,%f%%,%f,%d,%d,%d,%d" # per-class format
    else:
        fmt = "%-46s %7s %9s %6s %7s %7s %9s %7s %9s"
        fmtp = "%-46s %7d %9d %5.1f%% %7.2f %7d %9d %7d %9d" # per-class format
    fmts = "%46s %7d %8dk %5.1f%% %7.2f" # summary format
    print fmt % ("Class Name", "T.Count", "T.Bytes", "Pct", "AvgSize",
                 "C.Count", "C.Bytes", "O.Count", "O.Bytes")
    if not csv:
        print fmt % ('-'*46, '-'*7, '-'*9, '-'*5, '-'*7, '-'*7, '-'*9, '-'*7, '-'*9)
    typemap = rep.TYPEMAP.keys()
    typemap.sort(key=lambda a:rep.TYPESIZE[a])
    cumpct = 0.0
    for t in typemap:
        pct = rep.TYPESIZE[t] * 100.0 / rep.DBYTES
        cumpct += pct
        if csv:
            t_display = t
        else:
            t_display = shorten(t, 46)
        print fmtp % (t_display, rep.TYPEMAP[t], rep.TYPESIZE[t],
                      pct, rep.TYPESIZE[t] * 1.0 / rep.TYPEMAP[t],
                      rep.COIDSMAP[t], rep.CBYTESMAP[t],
                      rep.FOIDSMAP.get(t, 0), rep.FBYTESMAP.get(t, 0))

    if csv:
        return

    print fmt % ('='*46, '='*7, '='*9, '='*5, '='*7, '='*7, '='*9, '='*7, '='*9)
    print "%46s %7d %9s %6s %6.2fk" % ('Total Transactions', rep.TIDS, ' ',
        ' ', rep.DBYTES * 1.0 / rep.TIDS / 1024.0)
    print fmts % ('Total Records', rep.OIDS, rep.DBYTES / 1024.0, cumpct,
                  rep.DBYTES * 1.0 / rep.OIDS)

    print fmts % ('Current Objects', rep.COIDS, rep.CBYTES / 1024.0,
                  rep.CBYTES * 100.0 / rep.DBYTES,
                  rep.CBYTES * 1.0 / rep.COIDS)
    if rep.FOIDS:
        print fmts % ('Old Objects', rep.FOIDS, rep.FBYTES / 1024.0,
                      rep.FBYTES * 100.0 / rep.DBYTES,
                      rep.FBYTES * 1.0 / rep.FOIDS)

def analyze(path, use_dbm):
    fs = FileStorage(path, read_only=1)
    fsi = fs.iterator()
    report = Report(use_dbm)
    for txn in fsi:
        analyze_trans(report, txn)
    if use_dbm:
        shutil.rmtree(report.temp_dir)
    return report

def analyze_trans(report, txn):
    report.TIDS += 1
    for rec in txn:
        analyze_rec(report, rec)

def get_type(record):
    mod, klass = get_pickle_metadata(record.data)
    return "%s.%s" % (mod, klass)

def analyze_rec(report, record):
    oid = record.oid
    report.OIDS += 1
    if record.data is None:
        # No pickle -- aborted version or undo of object creation.
        return
    try:
        size = len(record.data) # Ignores various overhead
        report.DBYTES += size
        if oid not in report.OIDMAP:
            type = get_type(record)
            report.OIDMAP[oid] = type
            if report.use_dbm:
                report.USEDMAP[oid] = str(size)
            else:
                report.USEDMAP[oid] = size
            report.COIDS += 1
            report.CBYTES += size
            report.COIDSMAP[type] = report.COIDSMAP.get(type, 0) + 1
            report.CBYTESMAP[type] = report.CBYTESMAP.get(type, 0) + size
        else:
            type = report.OIDMAP[oid]
            if report.use_dbm:
                fsize = int(report.USEDMAP[oid])
                report.USEDMAP[oid] = str(size)
            else:
                fsize = report.USEDMAP[oid]
                report.USEDMAP[oid] = size
            report.FOIDS += 1
            report.FBYTES += fsize
            report.CBYTES += size - fsize
            report.FOIDSMAP[type] = report.FOIDSMAP.get(type, 0) + 1
            report.FBYTESMAP[type] = report.FBYTESMAP.get(type, 0) + fsize
            report.CBYTESMAP[type] = report.CBYTESMAP.get(type, 0) + size - fsize
        report.TYPEMAP[type] = report.TYPEMAP.get(type, 0) + 1
        report.TYPESIZE[type] = report.TYPESIZE.get(type, 0) + size
    except Exception, err:
        print err

__doc__ = """%(program)s: Data.fs analyzer

usage: %(program)s [options] /path/to/Data.fs

Options:
  -h, --help                 this help screen
  -c, --csv                  output CSV
  -d, --dbm                  use DBM as temporary storage to limit memory usage
"""

def usage(stream, msg=None):
    if msg:
        print >>stream, msg
        print >>stream
    program = os.path.basename(sys.argv[0])
    print >>stream, __doc__ % {"program": program}


def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:],
                                   'hcd', ['help', 'csv', 'dbm'])
        path = args[0]
    except (getopt.GetoptError, IndexError), msg:
        usage(sys.stderr, msg)
        sys.exit(2)
    csv = False
    use_dbm = False
    for opt, args in opts:
        if opt in ('-c', '--csv'):
            csv = True
        if opt in ('-d', '--dbm'):
            use_dbm = True
        if opt in ('-h', '--help'):
            usage(sys.stdout)
            sys.exit()
    report(analyze(path, use_dbm), csv)

if __name__ == "__main__":
    main()



##############################################################################
#
# Copyright (c) 2011 Nexedi SA and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly advised to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################
import os
import pkg_resources
import slapos.slap
import subprocess
import time
import xml_marshaller
import shutil
import sys
import glob
import argparse
import json
from slapos import client

MAX_PARTIONS = 10
MAX_SR_RETRIES = 3

def createFolder(folder, clean=False):
  if clean and os.path.exists(folder):
    shutil.rmtree(folder)
  if not(os.path.exists(folder)):
    os.mkdir(folder)

def createFolders(folder):
  if not(os.path.exists(folder)):
    os.makedirs(folder)

def isDir(folder):
  return os.path.isdir(folder)

def createFile(path, mode, content):
  f = open(path, mode)
  if os.path.exists(path):
    f.write(content)
    f.close()
  else:
    # error
    pass



class SlapOSControler(object):

  def __init__(self, working_directory, config, log):
    self.config = config
    self.software_root = os.path.join(working_directory, 'soft')
    self.instance_root = os.path.join(working_directory, 'inst')
    self.slapos_config = os.path.join(working_directory, 'slapos.cfg')
    self.log = log
    self.proxy_database = os.path.join(working_directory, 'proxy.db')
    self.instance_config = {}

  #TODO: implement a method to get all instance related the slapOS account
  # and deleting all old instances (based on creation date or name etc...)

  def createSlaposConfigurationFileAccount(self, key, certificate, slapos_url, config):
    # Create "slapos_account" directory in the "slapos_directory"
    slapos_account_directory = os.path.join(config['slapos_directory'], "slapos_account")
    createFolder(slapos_account_directory)
    # Create slapos-account files
    slapos_account_key_path = os.path.join(slapos_account_directory, "key")
    slapos_account_certificate_path = os.path.join(slapos_account_directory, "certificate")
    configuration_file_path = os.path.join(slapos_account_directory, "slapos.cfg")
    configuration_file_value = "[slapos]\nmaster_url = %s\n\
[slapconsole]\ncert_file = %s\nkey_file = %s" %(
                                  slapos_url,
                                  slapos_account_certificate_path,
                                  slapos_account_key_path)
    createFile(slapos_account_key_path, "w", key)
    createFile(slapos_account_certificate_path, "w", certificate)
    createFile(configuration_file_path, "w", configuration_file_value)
    self.configuration_file_path = configuration_file_path
    return slapos_account_key_path, slapos_account_certificate_path, configuration_file_path

  def supply(self, software_url, computer_id, state="available"):
    """
    Request the installation of a software release on a specific node
    Ex :
    my_controler.supply('kvm.cfg', 'COMP-726')
    """
    self.log('SlapOSControler : supply')
    parser = argparse.ArgumentParser()
    parser.add_argument("configuration_file")
    parser.add_argument("software_url")
    parser.add_argument("node")
    if os.path.exists(self.configuration_file_path):
      args = parser.parse_args([self.configuration_file_path, software_url, computer_id])
      config = client.Config()
      config.setConfig(args, args.configuration_file)
      try:
        local = client.init(config)
        local['supply'](software_url, computer_guid=computer_id, state=state)
        self.log('SlapOSControler : supply %s %s %s' %(software_url, computer_id, state))
      except:
        self.log("SlapOSControler.supply, \
                 exception in registerOpenOrder", exc_info=sys.exc_info())
        raise ValueError("Unable to supply (or remove)")
    else:
      raise ValueError("Configuration file not found.")

  def destroy(self, software_url, computer_id):
    """
    Request Deletetion of a software release on a specific node
    Ex :
    my_controler.destroy('kvm.cfg', 'COMP-726')
    """
    self.supply(self, software_url, computer_id, state="destroyed")
    
  def getInstanceRequestedState(self, reference):
    try:
      return self.instance_config[reference]['requested_state']
    except:
      raise ValueError("Instance '%s' not exist" %self.instance_config[reference])
      
  def request(self, reference, software_url, software_type=None,
            software_configuration=None, computer_guid=None, state='started'):
    """
    configuration_file_path (slapos acount)
    reference : instance title
    software_url : software path/url
    software_type : scalability
    software_configuration : dict { "_" : "{'toto' : 'titi'}" } 

    Ex :
    my_controler._request('Instance16h34Ben',
                               'kvm.cfg', 'cluster', { "_" : "{'toto' : 'titi'}" } )

    """
    self.log('SlapOSControler : request-->SlapOSMaster')
    current_intance_config = {'software_type':software_type,
                              'software_configuration':software_configuration,
                              'computer_guid':computer_guid,
                              'software_url':software_url,
                              'requested_state':state,
                              'partition':None
                              }
    self.instance_config[reference] = current_intance_config

    filter_kw = None
    if computer_guid != None:
      filter_kw = { "computer_guid": computer_guid }
    if os.path.exists(self.configuration_file_path):
      parser = argparse.ArgumentParser()
      parser.add_argument("configuration_file")
      args = parser.parse_args([self.configuration_file_path])
      config = client.Config()
      config.setConfig(args, args.configuration_file)
      try:
        local = client.init(config)
        partition = local['request'](
          software_release = software_url,
          partition_reference = reference,
          partition_parameter_kw = software_configuration,
          software_type = software_type,
          filter_kw = filter_kw,
          state = state)
        self.instance_config[reference]['partition'] = partition
        if state == 'destroyed':
          del self.instance_config[reference]
        if state == 'started':
          self.log('Instance started with configuration: %s' %str(software_configuration))
      except:
        self.log("SlapOSControler.request, \
              exception in registerOpenOrder", exc_info=sys.exc_info())
        raise ValueError("Unable to do this request")
    else:
      raise ValueError("Configuration file not found.")

  def _requestSpecificState(self, reference, state):
    self.request(reference,
        self.instance_config[reference]['software_url'],
        self.instance_config[reference]['software_type'],
        self.instance_config[reference]['software_configuration'],
        self.instance_config[reference]['computer_guid'],
        state=state
    )    
  
  def destroyInstance(self, reference):
    self.log('SlapOSControler : delete instance')
    try:
      self._requestSpecificState(reference, 'destroyed')
    except:
      raise ValueError("Can't delete instance '%s' (instance may not been created?)" %reference)
    
  def stopInstance(self, reference):
    self.log('SlapOSControler : stop instance')
    try:
      self._requestSpecificState(reference, 'stopped')
    except:
      raise ValueError("Can't stop instance '%s' (instance may not been created?)" %reference)
  
  def startInstance(self, reference):
    self.log('SlapOSControler : start instance')
    try:
      self._requestSpecificState(reference, 'started')
    except:
      raise ValueError("Can't start instance '%s' (instance may not been created?)" %reference)

  def updateInstanceXML(self, reference, software_configuration):
    """
    Update the XML configuration of an instance
    # Request same instance with different parameters.
    """
    self.log('SlapOSControler : updateInstanceXML')
    self.log('SlapOSControler : updateInstanceXML will request same'
             'instance with new XML configuration...')

    try:
      self.request(reference,
        self.instance_config[reference]['software_url'],
        self.instance_config[reference]['software_type'],
        software_configuration,
        self.instance_config[reference]['computer_guid'],
        state='started'
      )
    except:
      raise ValueError("Can't update instance '%s' (may not exist?)" %reference)

  def _resetSoftware(self):
    self.log('SlapOSControler : GOING TO RESET ALL SOFTWARE : %r' %
             (self.software_root,))
    if os.path.exists(self.software_root):
      shutil.rmtree(self.software_root)
    os.mkdir(self.software_root)
    os.chmod(self.software_root, 0750)

  def initializeSlapOSControler(self, slapproxy_log=None, process_manager=None,
        reset_software=False, software_path_list=None):
    self.process_manager = process_manager
    self.software_path_list = software_path_list
    self.log('SlapOSControler, initialize, reset_software: %r' % reset_software)
    config = self.config
    slapos_config_dict = self.config.copy()
    slapos_config_dict.update(software_root=self.software_root,
                              instance_root=self.instance_root,
                              proxy_database=self.proxy_database)
    open(self.slapos_config, 'w').write(pkg_resources.resource_string(
         'erp5.util.testnode', 'template/slapos.cfg.in') %
           slapos_config_dict)
    createFolder(self.software_root)
    createFolder(self.instance_root)
    # By erasing everything, we make sure that we are able to "update"
    # existing profiles. This is quite dirty way to do updates...
    if os.path.exists(self.proxy_database):
      os.unlink(self.proxy_database)
    kwargs = dict(close_fds=True, preexec_fn=os.setsid)
    if slapproxy_log is not None:
      slapproxy_log_fp = open(slapproxy_log, 'w')
      kwargs['stdout'] = slapproxy_log_fp
      kwargs['stderr'] = slapproxy_log_fp
    proxy = subprocess.Popen([config['slapproxy_binary'],
      self.slapos_config], **kwargs)
    process_manager.process_pid_set.add(proxy.pid)
    # XXX: dirty, giving some time for proxy to being able to accept
    # connections
    time.sleep(10)
    try:
      slap = slapos.slap.slap()
      self.slap = slap
      self.slap.initializeConnection(config['master_url'])
      # register software profile
      for path in self.software_path_list:
        slap.registerSupply().supply(
            path,
            computer_guid=config['computer_id'])
      computer = slap.registerComputer(config['computer_id'])
    except:
        self.log("SlapOSControler.initializeSlapOSControler, \
                 exception in registerSupply", exc_info=sys.exc_info())
        raise ValueError("Unable to initializeSlapOSControler")
    # Reset all previously generated software if needed
    if reset_software:
      self._resetSoftware()
    instance_root = self.instance_root
    if os.path.exists(instance_root):
      # delete old paritions which may exists in order to not get its data
      # (ex. MySQL db content) from previous testnode's runs
      # In order to be able to change partition naming scheme, do this at
      # instance_root level (such change happened already, causing problems).
      shutil.rmtree(instance_root)
    if not(os.path.exists(instance_root)):
      os.mkdir(instance_root)
    for i in range(0, MAX_PARTIONS):
      # create partition and configure computer
      # XXX: at the moment all partitions do share same virtual interface address
      # this is not a problem as usually all services are on different ports
      partition_reference = '%s-%s' %(config['partition_reference'], i)
      partition_path = os.path.join(instance_root, partition_reference)
      if not(os.path.exists(partition_path)):
        os.mkdir(partition_path)
      os.chmod(partition_path, 0750)
      computer.updateConfiguration(xml_marshaller.xml_marshaller.dumps({
           'address': config['ipv4_address'],
           'instance_root': instance_root,
           'netmask': '255.255.255.255',
           'partition_list': [
             {'address_list': [{'addr': config['ipv4_address'],
                               'netmask': '255.255.255.255'},
                              {'addr': config['ipv6_address'],
                               'netmask': 'ffff:ffff:ffff::'},],
              'path': partition_path,
              'reference': partition_reference,
              'tap': {'name': partition_reference},}],
           'reference': config['computer_id'],
           'software_root': self.software_root}))

  def spawn(self, *args, **kw):
    return self.process_manager.spawn(*args, **kw)

  def runSoftwareRelease(self, config, environment):
    self.log("SlapOSControler.runSoftwareRelease")
    cpu_count = os.sysconf("SC_NPROCESSORS_ONLN")
    os.putenv('MAKEFLAGS', '-j%s' % cpu_count)
    os.environ['PATH'] = environment['PATH']
    # a SR may fail for number of reasons (incl. network failures)
    # so be tolerant and run it a few times before giving up
    for runs in range(0, MAX_SR_RETRIES):
      status_dict = self.spawn(config['slapgrid_software_binary'],
                 '-v', '-c', '--all',
                 self.slapos_config, raise_error_if_fail=False,
                 log_prefix='slapgrid_sr', get_output=False)
      if status_dict['status_code'] == 0:
        break
    return status_dict

  def runComputerPartition(self, config, environment,
                           stdout=None, stderr=None):
    self.log("SlapOSControler.runComputerPartition")
    # cloudooo-json is required but this is a hack which should be removed
    config['instance_dict']['cloudooo-json'] = "{}"
    # report-url, report-project and suite-url are required to seleniumrunner
    # instance. This is a hack which must be removed.
    config['instance_dict']['report-url'] = config.get("report-url", "")
    config['instance_dict']['report-project'] = config.get("report-project", "")
    config['instance_dict']['suite-url'] = config.get("suite-url", "")
    # XXX: Hack to minimize writes to storage holding MySQL databases.
    #      Note this is something we want for all test suites, so it would
    #      not be better to define this parameter on each test suite.
    # XXX: Also move here the number of test db to create, so that software
    #      release stop create ones by default.
    config['instance_dict']['_'] = json.dumps({"mariadb": {
      "relaxed-writes": True,
      "mariadb-relaxed-writes": True, # BBB
      "test-database-amount": 30,
      }})
    for path in self.software_path_list:
      try:
        self.slap.registerOpenOrder().request(path,
          partition_reference='testing partition %s' % \
            self.software_path_list.index(path),
          partition_parameter_kw=config['instance_dict'])
      except:
        self.log("SlapOSControler.runComputerPartition, \
                 exception in registerOpenOrder", exc_info=sys.exc_info())
        raise ValueError("Unable to registerOpenOrder")

    # try to run for all partitions as one partition may in theory request another one 
    # this not always is required but curently no way to know how "tree" of partitions
    # may "expand"
    sleep_time = 0
    for runs in range(0, MAX_PARTIONS):
      status_dict = self.spawn(config['slapgrid_partition_binary'], '-v', '-c',
                 self.slapos_config, raise_error_if_fail=False,
                 log_prefix='slapgrid_cp', get_output=False)
      self.log('slapgrid_cp status_dict : %r' % (status_dict,))
      if status_dict['status_code'] in (0,):
        break
    # some hack to handle promise issues (should be only one of the two
    # codes, but depending on slapos versions, we have inconsistent status
    if status_dict['status_code'] in (1,2):
      status_dict['status_code'] = 0
    return status_dict



##############################################################################
#
# Copyright (c) 2011 Nexedi SA and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly advised to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################
from datetime import datetime,timedelta
import os
import subprocess
import sys
import time
import glob
import SlapOSControler
import json
import time
import shutil
import logging
import string
import random
from ProcessManager import SubprocessError, ProcessManager, CancellationError
from subprocess import CalledProcessError
from NodeTestSuite import SlapOSInstance
from Updater import Updater
from erp5.util import taskdistribution

class UnitTestRunner():
  def __init__(self, testnode):
    self.testnode = testnode
    self.slapos_controler = SlapOSControler.SlapOSControler(
                                  self.testnode.working_directory,
                                  self.testnode.config,
                                  self.testnode.log)

  def _prepareSlapOS(self, working_directory, slapos_instance, log,
          create_partition=1, software_path_list=None, **kw):
    """
    Launch slapos to build software and partitions
    """
    slapproxy_log = os.path.join(self.testnode.config['log_directory'],
                                  'slapproxy.log')
    log('Configured slapproxy log to %r' % slapproxy_log)
    reset_software = slapos_instance.retry_software_count > 10
    if reset_software:
      slapos_instance.retry_software_count = 0
    log('testnode, retry_software_count : %r' % \
             slapos_instance.retry_software_count)
    self.slapos_controler.initializeSlapOSControler(slapproxy_log=slapproxy_log,
       process_manager=self.testnode.process_manager, reset_software=reset_software,
       software_path_list=software_path_list)
    self.testnode.process_manager.supervisord_pid_file = os.path.join(\
         self.slapos_controler.instance_root, 'var', 'run', 'supervisord.pid')
    method_list= ["runSoftwareRelease"]
    if create_partition:
      method_list.append("runComputerPartition")
    for method_name in method_list:
      slapos_method = getattr(self.slapos_controler, method_name)
      log("Before status_dict = slapos_method(...)")
      status_dict = slapos_method(self.testnode.config,
                                  environment=self.testnode.config['environment'],
                                 )
      log(status_dict)
      log("After status_dict = slapos_method(...)")
      if status_dict['status_code'] != 0:
         slapos_instance.retry = True
         slapos_instance.retry_software_count += 1
         raise SubprocessError(status_dict)
      else:
         slapos_instance.retry_software_count = 0
    return status_dict

  def prepareSlapOSForTestNode(self, test_node_slapos):
    """
    We will build slapos software needed by the testnode itself,
    like the building of selenium-runner by default
    """
    return self._prepareSlapOS(self.testnode.config['slapos_directory'],
              test_node_slapos, self.testnode.log, create_partition=0,
              software_path_list=self.testnode.config.get("software_list"))

  def prepareSlapOSForTestSuite(self, node_test_suite):
    """
    Build softwares needed by testsuites
    """
    log = self.testnode.log
    if log is None:
      log = self.testnode.log
    return self._prepareSlapOS(node_test_suite.working_directory,
              node_test_suite, log,
              software_path_list=[node_test_suite.custom_profile_path])

  def runTestSuite(self, node_test_suite, portal_url, log=None):
    config = self.testnode.config
    parameter_list = []
    run_test_suite_path_list = glob.glob("%s/*/bin/runTestSuite" % \
        self.slapos_controler.instance_root)
    if not len(run_test_suite_path_list):
      raise ValueError('No runTestSuite provided in installed partitions.')
    run_test_suite_path = run_test_suite_path_list[0]
    run_test_suite_revision = node_test_suite.revision
    # Deal with Shebang size limitation
    invocation_list = self.testnode._dealShebang(run_test_suite_path)
    invocation_list.extend([run_test_suite_path,
                           '--test_suite', node_test_suite.test_suite,
                           '--revision', node_test_suite.revision,
                           '--test_suite_title', node_test_suite.test_suite_title,
                           '--node_quantity', config['node_quantity'],
                           '--master_url', portal_url])
    firefox_bin_list = glob.glob("%s/soft/*/parts/firefox/firefox-slapos" % \
        config["slapos_directory"])
    if len(firefox_bin_list):
      parameter_list.append('--firefox_bin')
    xvfb_bin_list = glob.glob("%s/soft/*/parts/xserver/bin/Xvfb" % \
        config["slapos_directory"])
    if len(xvfb_bin_list):
      parameter_list.append('--xvfb_bin')
    supported_paramater_set = self.testnode.process_manager.getSupportedParameterSet(
                           run_test_suite_path, parameter_list)
    if '--firefox_bin' in supported_paramater_set:
      invocation_list.extend(["--firefox_bin", firefox_bin_list[0]])
    if '--xvfb_bin' in supported_paramater_set:
      invocation_list.extend(["--xvfb_bin", xvfb_bin_list[0]])
    # TODO : include testnode correction ( b111682f14890bf )
    if hasattr(node_test_suite,'additional_bt5_repository_id'):
      additional_bt5_path = os.path.join(
              node_test_suite.working_directory,
              node_test_suite.additional_bt5_repository_id)
      invocation_list.extend(["--bt5_path", additional_bt5_path])
    # From this point, test runner becomes responsible for updating test
    # result. We only do cleanup if the test runner itself is not able
    # to run.
    SlapOSControler.createFolder(node_test_suite.test_suite_directory,
                                 clean=True)
    self.testnode.process_manager.spawn(*invocation_list,
                          cwd=node_test_suite.test_suite_directory,
                          log_prefix='runTestSuite', get_output=False)

  def getRelativePathUsage(self):
    """
    Used by the method testnode.constructProfile() to know
    if the software.cfg have to use relative path or not.
    """
    return False



#!/usr/bin/env python

import argparse
import os
import shutil
import time
import sys
import multiprocessing
import subprocess
import signal
import errno
import json
import logging
import logging.handlers
import glob
import urlparse
import httplib
import base64
from erp5.util.benchmark.argument import ArgumentType
from erp5.util.benchmark.performance_tester import PerformanceTester
from erp5.util import taskdistribution
from erp5.util.testnode import Utils

MAX_INSTALLATION_TIME = 60*50
MAX_TESTING_TIME = 60
MAX_GETTING_CONNECTION_TIME = 60*5

class DummyLogger(object):
  def __init__(self, func):
    for name in ('trace', 'debug', 'info', 'warn', 'warning', 'error',
      'critical', 'fatal'):
       setattr(self, name, func)

def getConnection(erp5_url, log):
  """
  Return a connection with the erp5 instance.
  """
  start_time = time.time()
  count = 0
  while MAX_GETTING_CONNECTION_TIME > time.time()-start_time:
    try:
      count = count + 1
      parsed = urlparse.urlparse(erp5_url)
      host = "%s:%s" % (parsed.hostname, str(parsed.port))
      if parsed.scheme == 'https':
        return httplib.HTTPSConnection(host)
      elif parsed.scheme == 'http':
        return httplib.HTTPConnection(host)
      else:
        raise ValueError("Protocol not implemented")
    except:
      log("Can't get connection to %s, we will retry." %erp5_url)
      time.sleep(10)
  raise ValueError("Cannot get new connection after %d try (for %s s)" %(count, str(time.time()-start_time)))

def waitFor0PendingActivities(erp5_url, log):
  """
  Waiting while there are no pending activities on the erp5 instance.
  """
  start_time = time.time()
  parsed = urlparse.urlparse(erp5_url)
  user = parsed.username;
  password = parsed.password;
  header_dict = {'Authorization': 'Basic %s' % \
  base64.encodestring('%s:%s' % (user, password)).strip()}

  count = 0
  ok = False
  while MAX_INSTALLATION_TIME > time.time()-start_time and not ok:
    zope_connection = getConnection(erp5_url, log)
    try:
      count = count + 1
      zope_connection.request(
        'GET', '/erp5/portal_activities/getMessageList',
        headers=header_dict
      )
      result = zope_connection.getresponse()
      message_list_text = result.read()
      message_list = [s.strip() for s in message_list_text[1:-1].split(',')]
      if len(message_list)==0:
        log("There is no pending activities.")
        ok = True
      #Hack to do not take into account persistent Alarm_installMailServer acitivities
      if len(message_list)==1 :
        log("1 pending activity but ok.")
        ok = True

      log("There is %d pending activities" %len(message_list))
      time.sleep(5)
    except:
      time.sleep(5)
      log("Getting activities failed, retry.")

  if not ok:
    raise ValueError("Cannot waitFor0PendingActivities after %d try (for %s s)" %(count, str(time.time()-start_time)))


def getCreatedDocumentNumberFromERP5(erp5_url, log):
  """
  Get the number of created documents from erp5 instance.
  """
  log("count docs number from ERP5 instance")
  count_retry = 0
  parsed = urlparse.urlparse(erp5_url)
  user = 'zope'
  password = 'insecure'
  header_dict = {'Authorization': 'Basic %s' % \
  base64.encodestring('%s:%s' % (user, password)).strip()}
  zope_connection = getConnection(erp5_url, log)
  while count_retry < 100 :
    try:
      zope_connection.request(
        'GET', '/erp5/count_docs_scalability',
        headers=header_dict
      )
      result = zope_connection.getresponse()
      return int(result.read())
    except:
      log("retry..")
      count_retry += 1
      time.sleep(15)
  raise ValueError("Impossible to get number of docs from ERP5")


# XXX: This import is required, just to populate sys.modules['test_suite'].
# Even if it's not used in this file. Yuck.
import product.ERP5Type.tests.ERP5TypeTestSuite


from subprocess import call

LOG_FILE_PREFIX = "performance_tester_erp5"
# Duration of a test case
TEST_CASE_DURATION = 60
# Maximum limit of documents to create during a test case
MAX_DOCUMENTS = 100000

class ScalabilityTest(object):
  def __init__(self, data, test_result):
    self.__dict__ = {}
    self.__dict__.update(data)
    self.test_result = test_result

def doNothing(**kwargs):
  pass

def makeSuite(test_suite=None, log=doNothing, **kwargs):
  # BBB tests (plural form) is only checked for backward compatibility
  for k in sys.modules.keys():
    if k in ('tests', 'test',) or k.startswith('tests.') or k.startswith('test.'):
      del sys.modules[k]
  singular_succeed = True
  while True:
    module_name, class_name = ('%s.%s' % (singular_succeed and 'test' or 'tests',
                                          test_suite)).rsplit('.', 1)
    try:
      suite_class = getattr(__import__(module_name, None, None, [class_name]),
                            class_name)
    except (AttributeError, ImportError):
      if not singular_succeed:
        raise
      singular_succeed = False
    else:
      break
  suite = suite_class(max_instance_count=1, **kwargs)
  return suite


class ScalabilityLauncher(object):
  def __init__(self):
    # Parse arguments
    self.__argumentNamespace = self._parseArguments(argparse.ArgumentParser(
          description='Run ERP5 benchmarking scalability suites.'))
    # Create Logger
    log_path = os.path.join(self.__argumentNamespace.log_path,
                            "runScalabilityTestSuite.log")
    logger_format = '%(asctime)s %(name)-13s: %(levelname)-8s %(message)s'
    formatter = logging.Formatter(logger_format)
    logging.basicConfig(level=logging.INFO,
                     format=logger_format)
    logger = logging.getLogger('runScalabilityTestSuite')
    logger.addHandler(logging.NullHandler())
    file_handler = logging.handlers.RotatingFileHandler(
        filename=log_path,
        maxBytes=20000000, backupCount=4)
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)
    self.log = logger.info

    # Proxy to with erp5 master test_result
    self.test_result = taskdistribution.TestResultProxy(
                        self.__argumentNamespace.test_suite_master_url,
                        1.0, DummyLogger(self.log),
                        self.__argumentNamespace.test_result_path,
                        self.__argumentNamespace.node_title,
                        self.__argumentNamespace.revision
                      )
  @staticmethod
  def _addParserArguments(parser):
    # Mandatory arguments
    parser.add_argument('--erp5-url',
                        metavar='ERP5_URL',
                        help='Main url of ERP5 instance to test')

    parser.add_argument('--test-result-path',
                        metavar='ERP5_TEST_RESULT_PATH',
                        help='ERP5 relative path of the test result')

    parser.add_argument('--revision',
                        metavar='REVISION',
                        help='Revision of the test_suite')

    parser.add_argument('--test-suite',
                        metavar='TEST_SUITE',
                        help='Name of the test suite')

    parser.add_argument('--node-title',
                        metavar='NODE_TITLE',
                        help='Title of the testnode which is running this'
                              'launcher')

    parser.add_argument('--test-suite-master-url',
                        metavar='TEST_SUITE_MASTER_URL',
                        help='Url to connect to the ERP5 Master testsuite taskditributor')

    parser.add_argument('--log-path',
                        metavar='LOG_PATH',
                        help='Log Path')

    parser.add_argument('--erp5-location',
                        metavar='ERP5_LOCATION',
                        help='Path to erp5 depository')

    parser.add_argument('--runner-path',
                        metavar='Runner_PATH',
                        help='runner Path')

  @staticmethod
  def _checkParsedArguments(namespace):
    return namespace

  @staticmethod
  def _parseArguments(parser):
    ScalabilityLauncher._addParserArguments(parser)
    namespace = parser.parse_args()
    ScalabilityLauncher._checkParsedArguments(namespace)
    return namespace

  def moveLogs(self, folder_name):
    # Get file paths
    file_to_move_list = glob.glob(os.path.join(self.__argumentNamespace.log_path,
                                "%s*.csv" %LOG_FILE_PREFIX))
    file_to_move_list += glob.glob(os.path.join(self.__argumentNamespace.log_path,
                                "%s*.log" %LOG_FILE_PREFIX))
    # Create folder
    new_directory_path = os.path.join(self.__argumentNamespace.log_path,
                                folder_name)
    if not os.path.exists(new_directory_path): os.makedirs(new_directory_path)
    # Move files
    for file_to_move in file_to_move_list:
      shutil.move(file_to_move, new_directory_path)

  def getRunningTest(self):
    """
    Return a ScalabilityTest with current running test case informations,
    or None if no test_case ready
    """
    data = self.test_result.getRunningTestCase()
    if not data:
      return None
    decoded_data = Utils.deunicodeData(json.loads(data))
    return ScalabilityTest(decoded_data, self.test_result)

  def run(self):
    self.log("Scalability Launcher started, with:")
    self.log("Test suite master url: %s" %self.__argumentNamespace.test_suite_master_url)
    self.log("Test suite: %s" %self.__argumentNamespace.test_suite)
    self.log("Test result path: %s" %self.__argumentNamespace.test_result_path)
    self.log("Revision: %s" %self.__argumentNamespace.revision)
    self.log("Node title: %s" %self.__argumentNamespace.node_title)
    self.log("ERP5 url: %s" %self.__argumentNamespace.erp5_url)

    error_message_set, exit_status = set(), 0

    # Get suite informations
    suite = makeSuite(self.__argumentNamespace.test_suite, self.log)
    test_suite_list = suite.getTestList()

    # Main loop
    while True:

      # Loop for getting new test case
      current_test = self.getRunningTest()
      while not current_test:
        time.sleep(15)
        current_test = self.getRunningTest()
      self.log("Test Case %s going to be run." %(current_test.title))

      # Prepare configuration
      current_test_number = int(current_test.title)
      test_duration = suite.getTestDuration(current_test_number)
      benchmarks_path = os.path.join(self.__argumentNamespace.erp5_location, suite.getTestPath())
      user_file_full_path = os.path.join(self.__argumentNamespace.erp5_location, suite.getUsersFilePath())
      user_file_path = os.path.split(user_file_full_path)[0]
      user_file = os.path.split(user_file_full_path)[1]
      tester_path = self.__argumentNamespace.runner_path
      user_number = suite.getUserNumber(current_test_number)
      repetition = suite.getTestRepetition(current_test_number)

      self.log("user_number: %s" %str(user_number))
      self.log("test_duration: %s seconds" %str(test_duration))

      # Store the number of documents generated for each iteration
      document_number = []

      # Repeat the same test several times to accurate test result
      for i in range(1, repetition+1):
        self.log("Repetition: %d/%d" %(i, repetition))

        # Get the number of documents present before running the test.
        waitFor0PendingActivities(self.__argumentNamespace.erp5_url, self.log)
        previous_document_number = getCreatedDocumentNumberFromERP5(self.__argumentNamespace.erp5_url, self.log)
        self.log("previous_document_number: %d" %previous_document_number)

        # Generate commands to run
        command_list = []
        user_index = 0
        for test_suite in test_suite_list:
          command_list.append([tester_path,
               self.__argumentNamespace.erp5_url,
               str(user_number/len(test_suite_list)),
               test_suite,
               '--benchmark-path-list', benchmarks_path,
               '--users-file-path', user_file_path,
               '--users-file', user_file,
               '--filename-prefix', "%s_%s_repetition%d" %(LOG_FILE_PREFIX, current_test.title, i),
               '--report-directory', self.__argumentNamespace.log_path,
               '--repeat', "%s" %str(MAX_DOCUMENTS),
               '--max-errors', str(1000000),
               '--user-index', str(user_index),
          ])
          user_index += user_number/len(test_suite_list)

        # Launch commands
        tester_process_list = []
        for command in command_list:
          self.log("command: %s" %str(command))
          tester_process_list.append(subprocess.Popen(command))

        # Sleep
        time.sleep(test_duration)

        # Stop
        for tester_process in tester_process_list:
          tester_process.send_signal(signal.SIGINT)
          self.log("End signal sent to the tester.")

        # Count created documents
        # Wait for 0 pending activities before counting
        waitFor0PendingActivities(self.__argumentNamespace.erp5_url, self.log)
        current_document_number = getCreatedDocumentNumberFromERP5(self.__argumentNamespace.erp5_url, self.log)
        created_document_number = current_document_number - previous_document_number
        self.log("previous_document_number: %d" %previous_document_number)
        self.log("current_document_number: %d" %current_document_number)
        self.log("created_document_number: %d" %created_document_number)
        document_number.append(created_document_number)
        # Move csv/logs
        self.moveLogs(current_test.title)

      self.log("Test Case %s is finish" %(current_test.title))

      # Get the maximum as choice
      maximum = 0
      for i in range(0,len(document_number)):
        if document_number[i] > maximum:
          maximum = document_number[i]

      # Send results to ERP5 master
      retry_time = 2.0
      proxy = taskdistribution.ServerProxy(
                  self.__argumentNamespace.test_suite_master_url,
                  allow_none=True
              ).portal_task_distribution
      test_result_line_test = taskdistribution.TestResultLineProxy(
                                proxy, retry_time, self.log,
                                current_test.relative_path,
                                current_test.title
                              )
      results = "created docs=%d\n"\
                "duration=%d\n"\
                "number of tests=%d\n"\
                "number of users=%d\n"\
                "tests=%s\n"\
                %(
                  maximum,
                  test_duration,
                  len(test_suite_list),
                  (user_number/len(test_suite_list))*len(test_suite_list),
                  '_'.join(test_suite_list)
                )
      self.log("Results: %s" %results)
      test_result_line_test.stop(stdout=results,
                      test_count=len(test_suite_list),
                      duration=test_duration)
      self.log("Test Case Stopped")

    #
    error_message_set = None
    exit_status = 0
    return error_message_set, exit_status

def main():
  error_message_set, exit_status = ScalabilityLauncher().run()
  for error_message in error_message_set:
    print >>sys.stderr, "ERROR: %s" % error_message

  sys.exit(exit_status)



##############################################################################
#
# Copyright (c) 2011 Nexedi SA and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly advised to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################
from datetime import datetime,timedelta
import os
import subprocess
import sys
import time
import glob
import SlapOSControler
import json
import time
import shutil
import logging
import string
import random
import Utils

import traceback

from ProcessManager import SubprocessError, ProcessManager, CancellationError
from subprocess import CalledProcessError
from Updater import Updater
from NodeTestSuite import NodeTestSuite, SlapOSInstance
from ScalabilityTestRunner import ScalabilityTestRunner
from UnitTestRunner import UnitTestRunner
from erp5.util import taskdistribution


DEFAULT_SLEEP_TIMEOUT = 120 # time in seconds to sleep
MAX_LOG_TIME = 15 # time in days we should keep logs that we can see through
                  # httd
MAX_TEMP_TIME = 0.01 # time in days we should keep temp files
supervisord_pid_file = None

PROFILE_PATH_KEY = 'profile_path'

class DummyLogger(object):
  def __init__(self, func):
    for name in ('trace', 'debug', 'info', 'warn', 'warning', 'error',
      'critical', 'fatal'):
       setattr(self, name, func)

class TestNode(object):

  def __init__(self, log, config, max_log_time=MAX_LOG_TIME,
               max_temp_time=MAX_TEMP_TIME):
    self.testnode_log = log
    self.log = log
    self.config = config or {}
    self.process_manager = ProcessManager(log)
    self.working_directory = config['working_directory']
    self.node_test_suite_dict = {}
    self.file_handler = None
    self.max_log_time = max_log_time
    self.max_temp_time = max_temp_time
    self.url_access = "https://[0::0]:0123" # Ipv6 + port of the node


  def checkOldTestSuite(self,test_suite_data):
    config = self.config
    installed_reference_set = set(os.listdir(self.working_directory))
    wished_reference_set = set([x['test_suite_reference'] for x in test_suite_data])
    to_remove_reference_set = installed_reference_set.difference(
                                 wished_reference_set)
    for y in to_remove_reference_set:
      fpath = os.path.join(self.working_directory,y)
      self.delNodeTestSuite(y)
      self.log("testnode.checkOldTestSuite, DELETING : %r" % (fpath,))
      if os.path.isdir(fpath):
       shutil.rmtree(fpath)
      else:
       os.remove(fpath)
  
  def getNodeTestSuite(self, reference):
    node_test_suite = self.node_test_suite_dict.get(reference)
    if node_test_suite is None:
      node_test_suite = NodeTestSuite(reference)
      self.node_test_suite_dict[reference] = node_test_suite
    return node_test_suite

  def delNodeTestSuite(self, reference):
    if self.node_test_suite_dict.has_key(reference):
      self.node_test_suite_dict.pop(reference)

  def _dealShebang(self,run_test_suite_path):
    line = open(run_test_suite_path, 'r').readline()
    invocation_list = []
    if line[:2] == '#!':
      invocation_list = line[2:].split()
    return invocation_list

  def constructProfile(self, node_test_suite, test_type, use_relative_path=False):
    config = self.config
    profile_content = ''
    assert len(node_test_suite.vcs_repository_list), "we must have at least one repository"
    profile_path_count = 0
    profile_content_list = []
    for vcs_repository in node_test_suite.vcs_repository_list:
      url = vcs_repository['url']
      buildout_section_id = vcs_repository.get('buildout_section_id', None)
      repository_path = vcs_repository['repository_path']
      try:
        profile_path = vcs_repository[PROFILE_PATH_KEY]
      except KeyError:
        pass
      else:
        profile_path_count += 1
        if profile_path_count > 1:
          raise ValueError(PROFILE_PATH_KEY + ' defined more than once')

        # Absolute path to relative path
        software_config_path = os.path.join(repository_path, profile_path)
        if use_relative_path :
          from_path = os.path.join(self.working_directory,
                                    node_test_suite.reference)
          software_config_path = os.path.relpath(software_config_path, from_path)


        profile_content_list.append("""
[buildout]
extends = %(software_config_path)s
""" %  {'software_config_path': software_config_path})

      # Construct sections
      if not(buildout_section_id is None):
        # Absolute path to relative
        if use_relative_path:
          from_path = os.path.join(self.working_directory,
                                    node_test_suite.reference)
          repository_path = os.path.relpath(repository_path, from_path)

        if test_type=="ScalabilityTest":
#          updater = Updater(repository_path, git_binary=self.config['git_binary'],
#          branch = vcs_repository.get('branch','master'), log=self.log, process_manager=self.process_manager)
#          updater.checkout()
#          revision = updater.getRevision()[1]
          all_revision = node_test_suite.revision
          # from 'sec1=xx-azer,sec2=yy-qwer,..' to [[sec1,azer],[sec2,qwer],..]
          revision_list = [ [x.split('=')[0],x.split('=')[1].split('-')[1]] for x in all_revision.split(',') ]
          # from [[sec1,azer],[sec2,qwer],..] to {sec1:azer,sec2:qwer,..}
          revision_dict = {branch:revision for branch,revision in revision_list}
          # <obfuscated_url> word is modified by in runner.prepareSlapOSForTestSuite()
          profile_content_list.append("""
[%(buildout_section_id)s]
repository = <obfuscated_url>/%(buildout_section_id)s/%(buildout_section_id)s.git
revision = %(revision)s
ignore-ssl-certificate = true
""" %     {'buildout_section_id': buildout_section_id,
          'revision': revision_dict[buildout_section_id]})
        else:
          profile_content_list.append("""
[%(buildout_section_id)s]
repository = %(repository_path)s
branch = %(branch)s
""" %     {'buildout_section_id': buildout_section_id,
          'repository_path' : repository_path,
          'branch' : vcs_repository.get('branch','master')})
    if not profile_path_count:
      raise ValueError(PROFILE_PATH_KEY + ' not defined')
    # Write file
    custom_profile = open(node_test_suite.custom_profile_path, 'w')
    # sort to have buildout section first
    profile_content_list.sort(key=lambda x: [x, ''][x.startswith('\n[buildout]')])
    custom_profile.write(''.join(profile_content_list))
    custom_profile.close()
    sys.path.append(repository_path)

  def getAndUpdateFullRevisionList(self, node_test_suite):
    full_revision_list = []
    config = self.config
    log = self.log
    for vcs_repository in node_test_suite.vcs_repository_list:
      repository_path = vcs_repository['repository_path']
      repository_id = vcs_repository['repository_id']
      branch = vcs_repository.get('branch')
      # Make sure we have local repository
      updater = Updater(repository_path, git_binary=config['git_binary'],
         branch=branch, log=log, process_manager=self.process_manager,
         working_directory=node_test_suite.working_directory,
         url=vcs_repository["url"])
      updater.checkout()
      revision = "-".join(updater.getRevision())
      full_revision_list.append('%s=%s' % (repository_id, revision))
    node_test_suite.revision = ','.join(full_revision_list)
    return full_revision_list

  def registerSuiteLog(self, test_result, node_test_suite):
    """
      Create a log dedicated for the test suite,
      and register the url to master node.
    """
    suite_log_path, folder_id = node_test_suite.createSuiteLog()
    self._initializeSuiteLog(suite_log_path)
    # TODO make the path into url
    test_result.reportStatus('LOG url', "%s/%s" % (self.config.get('httpd_url'),
                             folder_id), '')
    self.log("going to switch to log %r" % suite_log_path)
    self.process_manager.log = self.log = self.getSuiteLog()
    return suite_log_path

  def getSuiteLog(self):
    return self.suite_log

  def _initializeSuiteLog(self, suite_log_path):
    # remove previous handlers
    logger = logging.getLogger('testsuite')
    if self.file_handler is not None:
      logger.removeHandler(self.file_handler)
    # and replace it with new handler
    logger_format = '%(asctime)s %(name)-13s: %(levelname)-8s %(message)s'
    formatter = logging.Formatter(logger_format)
    logging.basicConfig(level=logging.INFO, format=logger_format)
    self.file_handler = logging.FileHandler(filename=suite_log_path)
    self.file_handler.setFormatter(formatter)
    logger.addHandler(self.file_handler)
    logger.info('Activated logfile %r output' % suite_log_path)
    self.suite_log = logger.info

  def checkRevision(self, test_result, node_test_suite):
    config = self.config
    log = self.log
    if log is None:
      log = self.log
    if node_test_suite.revision != test_result.revision:
     log('Disagreement on tested revision, checking out: %r' % (
          (node_test_suite.revision,test_result.revision),))
     for i, repository_revision in enumerate(test_result.revision.split(',')):
      vcs_repository = node_test_suite.vcs_repository_list[i]
      repository_path = vcs_repository['repository_path']
      revision = repository_revision.rsplit('-', 1)[1]
      # other testnodes on other boxes are already ready to test another
      # revision
      log('  %s at %s' % (repository_path, node_test_suite.revision))
      updater = Updater(repository_path, git_binary=config['git_binary'],
                        revision=revision, log=log,
                        process_manager=self.process_manager)
      updater.checkout()
      updater.git_update_server_info()
      updater.git_create_repository_link()
      node_test_suite.revision = test_result.revision

  def _cleanupLog(self):
    config = self.config
    log_directory = self.config['log_directory']
    now = time.time()
    for log_folder in os.listdir(log_directory):
      folder_path = os.path.join(log_directory, log_folder)
      if os.path.isdir(folder_path):
        if (now - os.stat(folder_path).st_mtime)/86400 > self.max_log_time:
          self.log("deleting log directory %r" % (folder_path,))
          shutil.rmtree(folder_path)

  def _cleanupTemporaryFiles(self):
    """
    buildout seems letting files under /tmp. To avoid regular error of
    missing disk space, remove old logs
    """
    temp_directory = self.config["system_temp_folder"]
    now = time.time()
    user_id = os.geteuid()
    for temp_folder in os.listdir(temp_directory):
      folder_path = os.path.join(temp_directory, temp_folder)
      if (temp_folder.startswith("tmp") or
          temp_folder.startswith("buildout")):
        try:
          stat = os.stat(folder_path)
          if stat.st_uid == user_id and \
              (now - stat.st_mtime)/86400 > self.max_temp_time:
            self.log("deleting temp directory %r" % (folder_path,))
            if os.path.isdir(folder_path):
              shutil.rmtree(folder_path)
            else:
              os.remove(folder_path)
        except OSError:
          self.log("_cleanupTemporaryFiles exception", exc_info=sys.exc_info())

  def cleanUp(self,test_result):
    log = self.log
    log('Testnode.cleanUp')
    self.process_manager.killPreviousRun()
    self._cleanupLog()
    self._cleanupTemporaryFiles()

  def run(self):
    log = self.log
    config = self.config
    slapgrid = None
    previous_revision_dict = {}
    revision_dict = {}
    test_result = None
    test_node_slapos = SlapOSInstance()
    test_node_slapos.edit(working_directory=self.config['slapos_directory'])
    try:
      while True:
        try:
          node_test_suite = None
          self.log = self.process_manager.log = self.testnode_log
          self.cleanUp(None)
          remote_test_result_needs_cleanup = False
          begin = time.time()
          portal_url = config['test_suite_master_url']
          portal = taskdistribution.TaskDistributionTool(portal_url,
                                                         logger=DummyLogger(log))
          self.portal = portal
          self.test_suite_portal = taskdistribution.TaskDistributor(
                                                        portal_url,
                                                        logger=DummyLogger(log))
          self.test_suite_portal.subscribeNode(node_title=config['test_node_title'],
                                               computer_guid=config['computer_id'])
          test_suite_data = self.test_suite_portal.startTestSuite(
                                               node_title=config['test_node_title'],
                                               computer_guid=config['computer_id'])
          if type(test_suite_data) == str:
            # Backward compatiblity
            test_suite_data = json.loads(test_suite_data)
          test_suite_data = Utils.deunicodeData(test_suite_data)
          log("Got following test suite data from master : %r" % \
              (test_suite_data,))
          try:
            my_test_type = self.test_suite_portal.getTestType()
          except:
            log("testnode, error during requesting getTestType() method \
from the distributor.")
            raise NotImplementedError
          # Select runner according to the test type
          if my_test_type == 'UnitTest':
            runner = UnitTestRunner(self)
          elif my_test_type == 'ScalabilityTest':
            runner = ScalabilityTestRunner(self)
          else:
            log("testnode, Runner type not implemented.", my_test_type)
            raise NotImplementedError
          log("Type of current test is %s" %(my_test_type,))
          # master testnode gets test_suites, slaves get nothing
          runner.prepareSlapOSForTestNode(test_node_slapos)
          # Clean-up test suites
          self.checkOldTestSuite(test_suite_data)
          for test_suite in test_suite_data:
            remote_test_result_needs_cleanup = False
            node_test_suite = self.getNodeTestSuite(
               test_suite["test_suite_reference"])
            node_test_suite.edit(
               working_directory=self.config['working_directory'],
               log_directory=self.config['log_directory'])
            node_test_suite.edit(**test_suite)
            # XXX: temporary hack to prevent empty test_suite
            if not hasattr(node_test_suite, 'test_suite'):
              node_test_suite.edit(test_suite='')
            run_software = True
            # kill processes from previous loop if any
            self.process_manager.killPreviousRun()
            self.getAndUpdateFullRevisionList(node_test_suite)
            # Write our own software.cfg to use the local repository
            self.constructProfile(node_test_suite, my_test_type, 
                                  runner.getRelativePathUsage())
            # Make sure we have local repository
            test_result = portal.createTestResult(node_test_suite.revision, [],
                     config['test_node_title'], False,
                     node_test_suite.test_suite_title,
                     node_test_suite.project_title)
            remote_test_result_needs_cleanup = True
            log("testnode, test_result : %r" % (test_result, ))
            if test_result is not None:
              self.registerSuiteLog(test_result, node_test_suite)
              self.checkRevision(test_result,node_test_suite)
              node_test_suite.edit(test_result=test_result)
              # Now prepare the installation of SlapOS and create instance
              status_dict = runner.prepareSlapOSForTestSuite(node_test_suite)
              # Give some time so computer partitions may start
              # as partitions can be of any kind we have and likely will never have
              # a reliable way to check if they are up or not ...
              time.sleep(20)
              if my_test_type == 'UnitTest':
                runner.runTestSuite(node_test_suite, portal_url)
              elif my_test_type == 'ScalabilityTest':
                error_message = None
                # A problem is appeared during runTestSuite
                if status_dict['status_code'] == 1:
                  error_message = "Software installation too long or error(s) are present during SR install."
                else:
                  status_dict = runner.runTestSuite(node_test_suite, portal_url)
                  # A problem is appeared during runTestSuite
                  if status_dict['status_code'] == 1:
                    error_message = status_dict['error_message']

                # If an error is appeared
                if error_message:
                  test_result.reportFailure(
                      stdout=error_message
                  )
                  self.log(error_message)
                  raise ValueError(error_message)
              else:
                raise NotImplementedError
                  
              # break the loop to get latest priorities from master
              break
            self.cleanUp(test_result)
        except (SubprocessError, CalledProcessError) as e:
          log("SubprocessError", exc_info=sys.exc_info())
          if remote_test_result_needs_cleanup:
            status_dict = e.status_dict or {}
            test_result.reportFailure(
              command=status_dict.get('command'),
              stdout=status_dict.get('stdout'),
              stderr=status_dict.get('stderr'),
            )
          continue
        except ValueError as e:
          # This could at least happens if runTestSuite is not found
          log("ValueError", exc_info=sys.exc_info())
          if node_test_suite is not None:
            node_test_suite.retry_software_count += 1
        except CancellationError, e:
          log("CancellationError", exc_info=sys.exc_info())
          self.process_manager.under_cancellation = False
          node_test_suite.retry = True
          continue
        except:
          ex_type, ex, tb = sys.exc_info()
          traceback.print_tb(tb)
          log("erp5testnode exception", exc_info=sys.exc_info())
          raise
        now = time.time()
        self.cleanUp(test_result)
        if (now-begin) < 120:
          sleep_time = 120 - (now-begin)
          log("End of processing, going to sleep %s" % sleep_time)
          time.sleep(sleep_time)
    except:
      log("Exception in error handling", exc_info=sys.exc_info())
    finally:
      if 'tb' in locals():
        del tb
      # Nice way to kill *everything* generated by run process -- process
      # groups working only in POSIX compilant systems
      # Exceptions are swallowed during cleanup phase
      log("GENERAL EXCEPTION, QUITING")
      self.cleanUp(test_result)
      log("GENERAL EXCEPTION, QUITING, cleanup finished")



import unittest
from unittest import TestCase

from erp5.util.testnode.testnode import TestNode
from erp5.util.testnode.NodeTestSuite import SlapOSInstance, NodeTestSuite
from erp5.util.testnode.ProcessManager import ProcessManager, SubprocessError
from erp5.util.testnode.Updater import Updater

from erp5.util.testnode.SlapOSMasterCommunicator import SlapOSMasterCommunicator
from erp5.util.testnode.SlapOSControler import SlapOSControler
from erp5.util.testnode.UnitTestRunner import UnitTestRunner
from erp5.util.testnode.ScalabilityTestRunner import ScalabilityTestRunner
from erp5.util.testnode.SlapOSControler import createFolder

from erp5.util.taskdistribution import TaskDistributor
from erp5.util.taskdistribution import TaskDistributionTool
from erp5.util.taskdistribution import TestResultProxy
import os
import shutil
import subprocess
import tempfile
import json
import time
import types
import re

class ERP5TestNode(TestCase):

  def setUp(self):
    self._temp_dir = tempfile.mkdtemp()
    self.working_directory = os.path.join(self._temp_dir, 'testnode')
    self.slapos_directory = os.path.join(self._temp_dir, 'slapos')
    self.software_directory = os.path.join(self._temp_dir, 'software_directory')
    self.test_suite_directory = os.path.join(self._temp_dir,'test_suite')
    self.environment = os.path.join(self._temp_dir,'environment')
    self.log_directory = os.path.join(self._temp_dir,'var/log/testnode')
    self.log_file = os.path.join(self.log_directory,'test.log')
    self.remote_repository0 = os.path.join(self._temp_dir, 'rep0')
    self.remote_repository1 = os.path.join(self._temp_dir, 'rep1')
    self.remote_repository2 = os.path.join(self._temp_dir, 'rep2')
    self.system_temp_folder = os.path.join(self._temp_dir,'tmp')
    os.mkdir(self.working_directory)
    os.mkdir(self.slapos_directory)
    os.mkdir(self.test_suite_directory)
    os.mkdir(self.software_directory)
    os.mkdir(self.environment)
    os.mkdir(self.system_temp_folder)
    os.makedirs(self.log_directory)
    os.close(os.open(self.log_file,os.O_CREAT))
    os.mkdir(self.remote_repository0)
    os.mkdir(self.remote_repository1)
    os.mkdir(self.remote_repository2)
    def log(*args,**kw):
      for arg in args:
        print "TESTNODE LOG : %r, %r" % (arg, kw)
    self.log = log

  def returnGoodClassRunner(self, test_type):
      if test_type == 'UnitTest':
        return UnitTestRunner
      elif test_type == 'ScalabilityTest':
        return ScalabilityTestRunner
      else:
        raise NotImplementedError

  def tearDown(self):
    shutil.rmtree(self._temp_dir, True)

  def getTestNode(self):
    # XXX how to get property the git path ?
    config = {}
    config["git_binary"] = "git"
    config["slapos_directory"] = self.slapos_directory
    config["working_directory"] = self.working_directory
    config["software_directory"] = self.software_directory
    config["node_quantity"] = 3
    config["test_suite_directory"] = self.test_suite_directory
    config["environment"] = self.environment
    config["log_directory"] = self.log_directory
    config["log_file"] = self.log_file
    config["test_suite_master_url"] = None
    config["hateoas_slapos_master_url"] = None
    config["test_node_title"] = "Foo-Test-Node"
    config["system_temp_folder"] = self.system_temp_folder
    config["computer_id"] = "COMP-TEST"
    config["server_url"] = "http://foo.bar"
    config["httpd_ip"] = "ff:ff:ff:ff:ff:ff:ff:ff"
    config["httpd_software_access_port"] = "9080"
    
    return TestNode(self.log, config)

  def getTestSuiteData(self, add_third_repository=False, reference="foo"):
    data = [{
       "test_suite": "Foo",
       "project_title": reference,
       "test_suite_title": "Foo-Test",
       "test_suite_reference": reference,
       "vcs_repository_list": [
            {'url': self.remote_repository0,
             'profile_path': 'software.cfg',
             'branch': 'master'},
            {'url': self.remote_repository1,
             'buildout_section_id': 'rep1',
             'branch': 'master'}]}]
    if add_third_repository:
      # add a third repository
      # insert in position zero since we already had bug when the profile_path
      # was defined in non-zero position when generating the profile
      data[0]['vcs_repository_list'].insert(0,
            {'url': self.remote_repository2,
             'buildout_section_id': 'rep2',
             'branch': 'foo'})
    return data

  def updateNodeTestSuiteData(self, node_test_suite,
                              add_third_repository=False):
    """
    Update from zero/Regenerate the testsuite
    """
    node_test_suite.edit(working_directory=self.working_directory,
       **self.getTestSuiteData(add_third_repository=add_third_repository)[0])

  def getCaller(self, **kw):
    class Caller(object):

      def __init__(self, **kw):
        self.__dict__.update(**kw)

      def __call__(self, command):
        return subprocess.check_output(command, **self.__dict__)
    return Caller(**kw)

  def generateTestRepositoryList(self, add_third_repository=False):
    commit_dict = {}
    repository_list = [self.remote_repository0, self.remote_repository1]
    if add_third_repository:
      repository_list.append(self.remote_repository2)
    for i, repository_path in enumerate(repository_list):
      call = self.getCaller(cwd=repository_path)
      call("git init".split())
      git_config = open(os.path.join(repository_path, '.git', 'config'), 'a')
      git_config.write("""
[user]
  name = a b
  email = a@b.c
""")
      git_config.close()
      call("touch first_file".split())
      call("git add first_file".split())
      call("git commit -v -m first_commit".split())
      my_file = open(os.path.join(repository_path, 'first_file'), 'w')
      my_file.write("initial_content%i" % i)
      my_file.close()
      call("git commit -av -m next_commit".split())
      output = call(['git', 'log', '--format=%H %s'])
      output = output.strip()
      output_line_list = output.split("\n")
      self.assertEquals(2, len(output_line_list))
      expected_commit_subject_list = ["next_commit", "first_commit"]
      commit_subject_list = [x.split()[1] for x in output_line_list]
      self.assertEquals(expected_commit_subject_list, commit_subject_list)
      commit_dict['rep%i' % i] = [x.split() for x in output_line_list]
      if repository_path == self.remote_repository2:
        output = call('git checkout master -b foo'.split())
    # commit_dict looks like
    # {'rep1': [['6669613db7239c0b7f6e1fdb82af6f583dcb3a94', 'next_commit'],
    #           ['4f1d14de1b04b4f878a442ee859791fa337bcf85', 'first_commit']],
    #  'rep0': [['fb2a61882148d705fd10ecd87278b458a59920a9', 'next_commit'],
    #           ['4f1d14de1b04b4f878a442ee859791fa337bcf85', 'first_commit']]}
    return commit_dict

  def test_01_getDelNodeTestSuite(self):
    """
    We should be able to get/delete NodeTestSuite objects inside test_node
    """
    test_node = self.getTestNode()
    node_test_suite = test_node.getNodeTestSuite('foo')
    self.assertEquals(0, node_test_suite.retry_software_count)
    node_test_suite.retry_software_count = 2
    self.assertEquals(2, node_test_suite.retry_software_count)
    node_test_suite = test_node.delNodeTestSuite('foo')
    node_test_suite = test_node.getNodeTestSuite('foo')
    self.assertEquals(0, node_test_suite.retry_software_count)

  def test_02_NodeTestSuiteWorkingDirectory(self):
    """
    Make sure we extend the working path with the node_test_suite reference
    """
    test_node = self.getTestNode()
    node_test_suite = test_node.getNodeTestSuite('foo')
    node_test_suite.edit(working_directory=self.working_directory)
    self.assertEquals("%s/foo" % self.working_directory,
                      node_test_suite.working_directory)
    self.assertEquals("%s/foo/test_suite" % self.working_directory,
                      node_test_suite.test_suite_directory)

  def test_03_NodeTestSuiteCheckDataAfterEdit(self):
    """
    When a NodeTestSuite instance is edited, the method _checkData
    analyse properties and add new ones
    """
    test_node = self.getTestNode()
    node_test_suite = test_node.getNodeTestSuite('foo')
    self.updateNodeTestSuiteData(node_test_suite)
    self.assertEquals(2, len(node_test_suite.vcs_repository_list))
    repository_path_list = []
    for vcs_repository in node_test_suite.vcs_repository_list:
      repository_path_list.append(vcs_repository['repository_path'])
    expected_list = ["%s/rep0" % node_test_suite.working_directory,
                     "%s/rep1" % node_test_suite.working_directory]
    self.assertEquals(expected_list, repository_path_list)

  def test_04_constructProfile(self, my_test_type='UnitTest'):
    """
    Check if the software profile is correctly generated
    """  
    test_node = self.getTestNode()
    test_node.test_suite_portal = TaskDistributor
    test_node.test_suite_portal.getTestNode = TaskDistributor.getTestType
    node_test_suite = test_node.getNodeTestSuite('foo')
    self.updateNodeTestSuiteData(node_test_suite, add_third_repository=True)
    node_test_suite.revision = 'rep1=1234-azerty,rep2=3456-qwerty'
    test_node.constructProfile(node_test_suite,my_test_type)
    self.assertEquals("%s/software.cfg" % (node_test_suite.working_directory,),
                      node_test_suite.custom_profile_path)
    profile = open(node_test_suite.custom_profile_path, 'r')
    if my_test_type=='UnitTest':
      expected_profile = """
[buildout]
extends = %(temp_dir)s/testnode/foo/rep0/software.cfg

[rep1]
repository = %(temp_dir)s/testnode/foo/rep1
branch = master

[rep2]
repository = %(temp_dir)s/testnode/foo/rep2
branch = foo
""" % {'temp_dir': self._temp_dir}
    else:
      revision1 = "azerty"
      revision2 = "qwerty"
      expected_profile = """
[buildout]
extends = %(temp_dir)s/testnode/foo/rep0/software.cfg

[rep1]
repository = <obfuscated_url>/rep1/rep1.git
revision = %(revision1)s
ignore-ssl-certificate = true

[rep2]
repository = <obfuscated_url>/rep2/rep2.git
revision = %(revision2)s
ignore-ssl-certificate = true
""" % {'temp_dir': self._temp_dir, 'revision1': revision1, 'revision2': revision2}
    self.assertEquals(expected_profile, profile.read())
    profile.close()

  def test_05_getAndUpdateFullRevisionList(self):
    """
    Check if we clone correctly repositories and get right revisions
    """
    commit_dict = self.generateTestRepositoryList()
    test_node = self.getTestNode()
    node_test_suite = test_node.getNodeTestSuite('foo')
    self.updateNodeTestSuiteData(node_test_suite)
    rev_list = test_node.getAndUpdateFullRevisionList(node_test_suite)
    self.assertEquals(2, len(rev_list))
    self.assertEquals(rev_list[0], 'rep0=2-%s' % commit_dict['rep0'][0][0])
    self.assertEquals(rev_list[1], 'rep1=2-%s' % commit_dict['rep1'][0][0])
    my_file = open(os.path.join(self.remote_repository1, 'first_file'), 'w')
    my_file.write("next_content")
    my_file.close()
    call = self.getCaller(cwd=self.remote_repository1)
    call("git commit -av -m new_commit".split())
    rev_list = test_node.getAndUpdateFullRevisionList(node_test_suite)
    self.assertTrue(rev_list[0].startswith('rep0=2-'))
    self.assertTrue(rev_list[1].startswith('rep1=3-'))
    self.assertEquals(2, len(node_test_suite.vcs_repository_list))
    for vcs_repository in node_test_suite.vcs_repository_list:
      self.assertTrue(os.path.exists(vcs_repository['repository_path']))

  def test_05b_changeRepositoryBranch(self, my_test_type='UnitTest'):
    """
    It could happen that the branch is changed for a repository. Testnode must
    be able to reset correctly the branch
    """
    commit_dict = self.generateTestRepositoryList(add_third_repository=True)
    test_node = self.getTestNode()
    node_test_suite = test_node.getNodeTestSuite('foo')
    self.updateNodeTestSuiteData(node_test_suite, add_third_repository=True)
    rev_list = test_node.getAndUpdateFullRevisionList(node_test_suite)
    self.assertEquals(3, len(rev_list))
    self.assertEquals(3, len(node_test_suite.vcs_repository_list))
    rep2_clone_path = [x['repository_path'] for x in \
                       node_test_suite.vcs_repository_list \
                       if x['repository_path'].endswith("rep2")][0]
    call = self.getCaller(cwd=rep2_clone_path)
    output = call("git branch".split()).strip()
    self.assertTrue("* foo" in output.split('\n'))
    vcs_repository_info = node_test_suite.vcs_repository_list[0]
    self.assertEquals(vcs_repository_info['repository_id'], 'rep2')
    self.assertEquals(vcs_repository_info['branch'], 'foo')
    # change it to master
    vcs_repository_info['branch'] = 'master'
    rev_list = test_node.getAndUpdateFullRevisionList(node_test_suite)
    output = call("git branch".split()).strip()
    print output
    self.assertTrue("* master" in output.split('\n'))
    # Add a third branch on remote, make sure we could switch to it
    remote_call = self.getCaller(cwd=self.remote_repository2)
    output = remote_call('git checkout master -b bar'.split())
    vcs_repository_info['branch'] = 'bar'
    rev_list = test_node.getAndUpdateFullRevisionList(node_test_suite)
    output = call("git branch".split()).strip()
    self.assertTrue("* bar" in output.split('\n'))

  def test_05c_changeRepositoryUrl(self):
    """
    It could happen that the url is changed for a repository (new place, or
    change of username and password). testnode must be able to erase and clone
    again the repository
    """
    commit_dict = self.generateTestRepositoryList(add_third_repository=True)
    test_node = self.getTestNode()
    node_test_suite = test_node.getNodeTestSuite('foo')
    self.updateNodeTestSuiteData(node_test_suite)
    rev_list = test_node.getAndUpdateFullRevisionList(node_test_suite)
    self.assertEquals(2, len(rev_list))
    self.assertEquals(2, len(node_test_suite.vcs_repository_list))
    # patch deleteRepository to make sure it will be called once for the wrong
    # repos, and not for the repos which has not changed
    deleted_repository_path_list = []
    original_deleteRepository = Updater.deleteRepository
    try:
      def deleteRepository(self):
        deleted_repository_path_list.append(self.repository_path)
        original_deleteRepository(self)
      Updater.deleteRepository = deleteRepository
      # change the url of the first repository
      vcs_repository_info = node_test_suite.vcs_repository_list[0]
      vcs_repository_info["url"] = self.remote_repository2
      rep0_clone_path = [x['repository_path'] for x in \
                         node_test_suite.vcs_repository_list \
                         if x['repository_path'].endswith("rep0")][0]
      call = self.getCaller(cwd=rep0_clone_path)
      self.assertEquals(call("git config --get remote.origin.url".split()).strip(),
                        self.remote_repository0)
      rev_list = test_node.getAndUpdateFullRevisionList(node_test_suite)
      self.assertEquals(call("git config --get remote.origin.url".split()).strip(),
                        self.remote_repository2)
      self.assertEquals([rep0_clone_path], deleted_repository_path_list)
    finally:
      Updater.deleteRepository = original_deleteRepository

  def test_06_checkRevision(self):
    """
    Check if we are able to restore older commit hash if master decide so
    """
    commit_dict = self.generateTestRepositoryList()
    test_node = self.getTestNode()
    node_test_suite = test_node.getNodeTestSuite('foo')
    self.updateNodeTestSuiteData(node_test_suite)
    rev_list = test_node.getAndUpdateFullRevisionList(node_test_suite)
    def getRepInfo(count=0, hash=0):
      assert count or hash
      info_list = []
      for vcs_repository in node_test_suite.vcs_repository_list:
        call = self.getCaller(cwd=vcs_repository['repository_path'])
        if count:
          info_list.append(
            call("git rev-list --topo-order --count HEAD".split()).strip())
        if hash:
          info_list.append(
            call("git log -n1 --format=%H".split()).strip())
      return info_list
    self.assertEquals(['2', '2'], getRepInfo(count=1))
    self.assertEquals([commit_dict['rep0'][0][0],commit_dict['rep1'][0][0]],
                      getRepInfo(hash=1))
    class TestResult(object):
      pass
    test_result = TestResult()
    # for test result to be one commit late for rep1 to force testnode to
    # reset tree to older version
    test_result.revision = 'rep0=2-%s,rep1=1-%s' % (commit_dict['rep0'][0][0],
                                                    commit_dict['rep1'][1][0])
    test_node.checkRevision(test_result, node_test_suite)
    expected_count_list = ['2', '1']
    self.assertEquals(['2', '1'], getRepInfo(count=1))
    self.assertEquals([commit_dict['rep0'][0][0],commit_dict['rep1'][1][0]],
                      getRepInfo(hash=1))

  def test_07_checkExistingTestSuite(self):
    test_node = self.getTestNode()
    test_suite_data = self.getTestSuiteData(add_third_repository=True)
    self.assertEquals([], os.listdir(self.working_directory))
    test_node.checkOldTestSuite(test_suite_data)
    self.assertEquals([], os.listdir(self.working_directory))
    os.mkdir(os.path.join(self.working_directory, 'foo'))
    self.assertEquals(['foo'], os.listdir(self.working_directory))
    test_node.checkOldTestSuite(test_suite_data)
    self.assertEquals(['foo'], os.listdir(self.working_directory))
    os.mkdir(os.path.join(self.working_directory, 'bar'))
    self.assertEquals(set(['bar','foo']),
                      set(os.listdir(self.working_directory)))
    test_node.checkOldTestSuite(test_suite_data)
    self.assertEquals(['foo'], os.listdir(self.working_directory))

  def test_08_getSupportedParamaterSet(self):
    original_spawn = ProcessManager.spawn
    try:
      def get_help(self, *args, **kw):
        return {'stdout': """My Program
                  --foo  foo
                  --bar  bar"""}
      ProcessManager.spawn = get_help
      process_manager = ProcessManager(log=None)
      parameter_list = ['--foo', '--baz']
      expected_suported_parameter_set = set(['--foo'])
      supported_parameter_set = process_manager.getSupportedParameterSet(
                                 "dummy_program_path", parameter_list)
      self.assertEquals(expected_suported_parameter_set, supported_parameter_set)
    finally:
      ProcessManager.spawn = original_spawn

  def test_09_runTestSuite(self, my_test_type='UnitTest'):
    """
    Check parameters passed to runTestSuite
    Also make sure that --firefox_bin and --xvfb_bin are passed when needed
    """
    original_getSupportedParameter = ProcessManager.getSupportedParameterSet
    original_spawn = ProcessManager.spawn
    try:
      # Create a file
      def _createPath(path_to_create, end_path):
        os.makedirs(path_to_create)
        return os.close(os.open(os.path.join(path_to_create,
                                 end_path),os.O_CREAT))
      
      def get_parameters(self, *args, **kw):
        call_parameter_list.append({'args': [x for x in args], 'kw':kw})
      
      def patch_getSupportedParameterSet(self, run_test_suite_path, parameter_list,):
       if '--firefox_bin' and '--xvfb_bin' in parameter_list:
         return set(['--firefox_bin','--xvfb_bin'])
       else:
         return []
      
      test_node = self.getTestNode()
      RunnerClass = self.returnGoodClassRunner(my_test_type)
      runner = RunnerClass(test_node)
      # Create and initialise/regenerate a nodetestsuite
      node_test_suite = test_node.getNodeTestSuite('foo')
      self.updateNodeTestSuiteData(node_test_suite)
      node_test_suite.revision = 'dummy'
      # Path to the dummy runable
      run_test_suite_path = _createPath(
          os.path.join(runner.slapos_controler.instance_root,'a/bin'),'runTestSuite')

      def checkRunTestSuiteParameters(additional_parameter_list=None):
        ProcessManager.getSupportedParameterSet = patch_getSupportedParameterSet
        ProcessManager.spawn = get_parameters
        RunnerClass = self.returnGoodClassRunner(my_test_type)
        runner = RunnerClass(test_node)
        runner.runTestSuite(node_test_suite,"http://foo.bar")
        expected_parameter_list = ['%s/a/bin/runTestSuite'
           %(runner.slapos_controler.instance_root), '--test_suite', 'Foo', '--revision',
           'dummy', '--test_suite_title', 'Foo-Test', '--node_quantity', 3, '--master_url',
           'http://foo.bar']
        if additional_parameter_list:
          expected_parameter_list.extend(additional_parameter_list)
        self.assertEqual(call_parameter_list[0]['args'], expected_parameter_list)

      call_parameter_list = []
      checkRunTestSuiteParameters()
      _createPath(os.path.join(test_node.config['slapos_directory'], 'soft/a/parts/firefox'),'firefox-slapos')
      _createPath(os.path.join(test_node.config['slapos_directory'], 'soft/a/parts/xserver/bin'),'Xvfb')
      call_parameter_list = []
      checkRunTestSuiteParameters(additional_parameter_list=['--firefox_bin',
        '%s/soft/a/parts/firefox/firefox-slapos'
         %(test_node.config['slapos_directory']),
        '--xvfb_bin',
        '%s/soft/a/parts/xserver/bin/Xvfb'
          %(test_node.config['slapos_directory'])])
    finally:
      ProcessManager.getSupportedParameterSet = original_getSupportedParameter
      ProcessManager.spawn = original_spawn

  def test_10_prepareSlapOS(self, my_test_type='UnitTest'):
    test_node = self.getTestNode()
    test_node_slapos = SlapOSInstance()
    RunnerClass = self.returnGoodClassRunner(my_test_type)
    runner = RunnerClass(test_node)
    node_test_suite = test_node.getNodeTestSuite('foo')
    node_test_suite.edit(working_directory=self.working_directory)
    status_dict = {"status_code" : 0}
    global call_list
    call_list = []
    class Patch:
      def __init__(self, method_name, status_code=0):
        self.method_name = method_name
        self.status_code = status_code
      def __call__(self, *args, **kw):
        global call_list
        call_list.append({"method_name": self.method_name,
                         "args": [x for x in args],
                          "kw": kw})
        return {"status_code": self.status_code}
    
    SlapOSControler.initializeSlapOSControler = Patch("initializeSlapOSControler")
    SlapOSControler.runSoftwareRelease = Patch("runSoftwareRelease")
    SlapOSControler.runComputerPartition = Patch("runComputerPartition")
    method_list_for_prepareSlapOSForTestNode = ["initializeSlapOSControler",
                                                   "runSoftwareRelease"]
    method_list_for_prepareSlapOSForTestSuite = ["initializeSlapOSControler",
                                 "runSoftwareRelease", "runComputerPartition"]
    runner.prepareSlapOSForTestNode(test_node_slapos)
    self.assertEquals(method_list_for_prepareSlapOSForTestNode,
                      [x["method_name"] for x in call_list])
    call_list = []
    runner.prepareSlapOSForTestSuite(node_test_suite)
    self.assertEquals(method_list_for_prepareSlapOSForTestSuite,
                      [x["method_name"] for x in call_list])
    call_list = []
    SlapOSControler.runSoftwareRelease = Patch("runSoftwareRelease", status_code=1)
    # TODO : write a test for scalability case
    self.assertRaises(SubprocessError, runner.prepareSlapOSForTestSuite,
                     node_test_suite)

  def test_11_run(self, my_test_type='UnitTest', grade='master'):
    def doNothing(self, *args, **kw):
        pass
    # Used in case of 'ScalabilityTest'
    def patch_getTestType(self, *args, **kw):
      return my_test_type
    def patch_getSlaposAccountKey(self, *args, **kw):
      return "key"
    def patch_getSlaposAccountCertificate(self, *args, **kw):
      return "Certificate"
    def patch_getSlaposUrl(self, *args, **kw):
      return "http://Foo"
    def patch_getSlaposHateoasUrl(self, *args, **kw):
      return "http://Foo"
    def patch_generateConfiguration(self, *args, **kw):
      return json.dumps({"configuration_list": [], "involved_nodes_computer_guid"\
: [], "error_message": "No error.", "launcher_nodes_computer_guid": [], \
"launchable": False, "randomized_path" : "azertyuiop"})
    def patch_isMasterTestnode(self, *args, **kw):
      return (grade == 'master')
    def patch_isHostingSubscriptionReady(self, *args, **kw):
      return True
    def patch_isRegisteredHostingSubscription(self, *args, **kw):
      return True      
    test_self = self
    test_result_path_root = os.path.join(test_self._temp_dir,'test/results')
    os.makedirs(test_result_path_root)
    global counter
    counter = 0
    def patch_startTestSuite(self,node_title,computer_guid='unknown'):
      global counter
      config_list = []
      # Sclalability slave testnode is not directly in charge of testsuites
      if my_test_type == 'ScalabilityTest' and grade == 'slave':
        if counter == 5:
          raise StopIteration
        counter += 1
        return json.dumps([])
          
      def _checkExistingTestSuite(reference_set):
        test_self.assertEquals(set(reference_set),
                  set(os.listdir(test_node.working_directory)))
        for x in reference_set:
          test_self.assertTrue(os.path.exists(os.path.join(
                             test_node.working_directory,x)),True)
      if counter == 0:
        config_list.append(test_self.getTestSuiteData(reference='foo')[0])
        config_list.append(test_self.getTestSuiteData(reference='bar')[0])
      elif counter == 1:
        _checkExistingTestSuite(set(['foo']))
        config_list.append(test_self.getTestSuiteData(reference='bar')[0])
        config_list.append(test_self.getTestSuiteData(reference='foo')[0])
      elif counter == 2:
        _checkExistingTestSuite(set(['foo','bar']))
        config_list.append(test_self.getTestSuiteData(reference='foo')[0])
        config_list.append(test_self.getTestSuiteData(reference='qux')[0])
      elif counter == 3:
        _checkExistingTestSuite(set(['foo','qux']))
        config_list.append(test_self.getTestSuiteData(reference='foox')[0])
      elif counter == 4:
        _checkExistingTestSuite(set(['foox']))
        config_list.append(test_self.getTestSuiteData(reference='bax')[0])
      elif counter == 5:
        _checkExistingTestSuite(set(['bax']))
        raise StopIteration
      counter += 1
      return json.dumps(config_list)
    def patch_createTestResult(self, revision, test_name_list, node_title,
            allow_restart=False, test_title=None, project_title=None):
      global counter
      # return no test to check if run method will run the next test suite
      if counter == 3 and project_title != 'qux':
        result = None
      else:
        test_result_path = os.path.join(test_result_path_root, test_title)
        result =  TestResultProxy(self._proxy, self._retry_time,
                self._logger, test_result_path, node_title, revision)
      return result
    def patch_runTestSuite(self, *argv, **kw):
      return {'status_code':0}
    original_sleep = time.sleep
    time.sleep = doNothing
    self.generateTestRepositoryList()
    RunnerClass = self.returnGoodClassRunner(my_test_type)
    # Patch
    if my_test_type == "ScalabilityTest":
      original_getSlaposAccountKey = TaskDistributor.getSlaposAccountKey
      original_getSlaposAccountCertificate = TaskDistributor.getSlaposAccountCertificate
      original_getSlaposUrl = TaskDistributor.getSlaposUrl
      original_getSlaposHateoasUrl = TaskDistributor.getSlaposHateoasUrl
      original_generateConfiguration = TaskDistributor.generateConfiguration
      original_isMasterTestnode = TaskDistributor.isMasterTestnode
      original_updateInstanceXML = RunnerClass._updateInstanceXML
      original_isHostingSubscriptionReady = SlapOSMasterCommunicator.isHostingSubscriptionReady
      original_isRegisteredHostingSubscription = SlapOSMasterCommunicator.isRegisteredHostingSubscription
      original_SlapOSMasterCommunicator__init__ = SlapOSMasterCommunicator.__init__
      TaskDistributor.getSlaposAccountKey = patch_getSlaposAccountKey
      TaskDistributor.getSlaposAccountCertificate = patch_getSlaposAccountCertificate
      TaskDistributor.getSlaposUrl = patch_getSlaposUrl
      TaskDistributor.getSlaposHateoasUrl = patch_getSlaposHateoasUrl
      TaskDistributor.generateConfiguration = patch_generateConfiguration
      TaskDistributor.isMasterTestnode = patch_isMasterTestnode
      RunnerClass._updateInstanceXML = doNothing
      SlapOSMasterCommunicator.isHostingSubscriptionReady = patch_isHostingSubscriptionReady
      SlapOSMasterCommunicator.isRegisteredHostingSubscription = patch_isRegisteredHostingSubscription
      SlapOSMasterCommunicator.__init__ = doNothing
    original_startTestSuite = TaskDistributor.startTestSuite
    original_subscribeNode = TaskDistributor.subscribeNode
    original_getTestType = TaskDistributor.getTestType
    original_createTestResult = TaskDistributionTool.createTestResult
    TaskDistributor.startTestSuite = patch_startTestSuite
    TaskDistributor.subscribeNode = doNothing
    TaskDistributor.getTestType = patch_getTestType
    TaskDistributionTool.createTestResult = patch_createTestResult

    # TestNode
    test_node = self.getTestNode()  
    # Modify class UnitTestRunner(or more after) method 
    original_prepareSlapOS = RunnerClass._prepareSlapOS
    original_runTestSuite = RunnerClass.runTestSuite
    RunnerClass._prepareSlapOS = doNothing
    RunnerClass.runTestSuite = patch_runTestSuite
    SlapOSControler.initializeSlapOSControler = doNothing
    # Inside test_node a runner is created using new UnitTestRunner methods
    test_node.run()
    self.assertEquals(5, counter)
    time.sleep = original_sleep
    # Restore old class methods
    if my_test_type == "ScalabilityTest":
      TaskDistributor.getSlaposAccountKey = original_getSlaposAccountKey
      TaskDistributor.getSlaposAccountCertificate = original_getSlaposAccountCertificate
      TaskDistributor.getSlaposUrl = original_getSlaposUrl
      TaskDistributor.getSlaposHateoasUrl = original_getSlaposHateoasUrl
      TaskDistributor.generateConfiguration = original_generateConfiguration
      TaskDistributor.isMasterTestnode = original_isMasterTestnode
      RunnerClass._updateInstanceXML = original_updateInstanceXML
      SlapOSMasterCommunicator.isHostingSubscriptionReady = original_isHostingSubscriptionReady
      SlapOSMasterCommunicator.isRegisteredHostingSubscription = original_isRegisteredHostingSubscription
      SlapOSMasterCommunicator.__init__ = original_SlapOSMasterCommunicator__init__
    TaskDistributor.startTestSuite = original_startTestSuite
    TaskDistributionTool.createTestResult = original_createTestResult
    TaskDistributionTool.subscribeNode = original_subscribeNode
    TaskDistributionTool.getTestType = original_getTestType
    RunnerClass._prepareSlapOS = original_prepareSlapOS
    RunnerClass.runTestSuite = original_runTestSuite

  def test_12_spawn(self):
    def _checkCorrectStatus(expected_status,*args):
      result = process_manager.spawn(*args)
      self.assertEqual(result['status_code'], expected_status)
    process_manager = ProcessManager(log=self.log, max_timeout=1)
    _checkCorrectStatus(0, *['sleep','0'])
    # We must make sure that if the command is too long that
    # it will be automatically killed
    self.assertRaises(SubprocessError, process_manager.spawn, 'sleep','3')

  def test_13_SlaposControlerResetSoftware(self):
    test_node = self.getTestNode()
    controler = SlapOSControler(self.working_directory,
                                test_node.config, self.log)
    os.mkdir(controler.software_root)
    file_name = 'AC_Ra\xc3\xadzertic\xc3\xa1ma'
    non_ascii_file = open(os.path.join(controler.software_root, file_name), 'w')
    non_ascii_file.close()
    self.assertEquals([file_name], os.listdir(controler.software_root))
    controler._resetSoftware()
    self.assertEquals([], os.listdir(controler.software_root))

  def test_14_createFolder(self):
    test_node = self.getTestNode()
    node_test_suite = test_node.getNodeTestSuite('foo')
    node_test_suite.edit(working_directory=self.working_directory)
    folder = node_test_suite.test_suite_directory
    self.assertEquals(False, os.path.exists(folder))
    createFolder(folder)
    self.assertEquals(True, os.path.exists(folder))
    to_drop_path = os.path.join(folder, 'drop')
    to_drop = open(to_drop_path, 'w')
    to_drop.close()
    self.assertEquals(True, os.path.exists(to_drop_path))
    createFolder(folder, clean=True)
    self.assertEquals(False, os.path.exists(to_drop_path))

  def test_15_suite_log_directory(self, my_test_type='UnitTest', grade='master'):
    def doNothing(self, *args, **kw):
      pass
    # Used in case of 'ScalabilityTest'
    def patch_getTestType(self, *args, **kw):
      return my_test_type
    def patch_getSlaposAccountKey(self, *args, **kw):
      return "key"
    def patch_getSlaposAccountCertificate(self, *args, **kw):
      return "Certificate"
    def patch_getSlaposUrl(self, *args, **kw):
      return "http://Foo"
      return "Certificate"
    def patch_getSlaposHateoasUrl(self, *args, **kw):
      return "http://Foo"
    def patch_generateConfiguration(self, *args, **kw):
      return json.dumps({"configuration_list": [], "involved_nodes_computer_guid"\
: [], "error_message": "No error.", "launcher_nodes_computer_guid": [], \
"launchable": False, "randomized_path" : "azertyuiop"})
    def patch_isMasterTestnode(self, *args, **kw):
      return grade == 'master'
    def patch_isHostingSubscriptionReady(self, *args, **kw):
      return True
    def patch_isRegisteredHostingSubscription(self, *args, **kw):
      return True     
    test_self = self
    test_result_path_root = os.path.join(test_self._temp_dir,'test/results')
    os.makedirs(test_result_path_root)
    global counter
    counter = 0
    def patch_startTestSuite(self,node_title,computer_guid='unknown'):
      global counter
      config_list = [test_self.getTestSuiteData(reference='aa')[0],
                     test_self.getTestSuiteData(reference='bb')[0]]
      if counter in (1, 2):
        config_list.reverse()
      elif counter == 3:
        raise StopIteration
      counter += 1
      return json.dumps(config_list)
    def patch_createTestResult(self, revision, test_name_list, node_title,
            allow_restart=False, test_title=None, project_title=None):
      test_result_path = os.path.join(test_result_path_root, test_title)
      result = TestResultProxy(self._proxy, self._retry_time,
               self._logger, test_result_path, node_title, revision)
      return result
    def patch_runTestSuite(self,*argv, **kw):
      return {'status_code':0}
    def checkTestSuite(test_node):
      test_node.node_test_suite_dict
      rand_part_set = set()
      self.assertEquals(2, len(test_node.node_test_suite_dict))
      assert(test_node.suite_log is not None)
      assert(isinstance(test_node.suite_log, types.MethodType))
      for ref, suite in test_node.node_test_suite_dict.items():
        self.assertTrue('var/log/testnode/%s' % suite.reference in \
                         suite.suite_log_path,
                         "Incorrect suite log path : %r" % suite.suite_log_path)
        assert(suite.suite_log_path.endswith('suite.log'))
        m = re.match('.*\-(.*)\/suite.log', suite.suite_log_path)
        rand_part = m.groups()[0]
        assert(len(rand_part) == 32)
        assert(rand_part not in rand_part_set)
        rand_part_set.add(rand_part)
        suite_log = open(suite.suite_log_path, 'r')
        self.assertEquals(1, len([x for x in suite_log.readlines() \
                              if x.find("Activated logfile")>=0]))

    RunnerClass = self.returnGoodClassRunner(my_test_type)
    original_sleep = time.sleep
    time.sleep = doNothing
    self.generateTestRepositoryList()
    if my_test_type == "ScalabilityTest":
      original_getSlaposAccountKey = TaskDistributor.getSlaposAccountKey
      original_getSlaposAccountCertificate = TaskDistributor.getSlaposAccountCertificate
      original_getSlaposUrl = TaskDistributor.getSlaposUrl
      original_getSlaposHateoasUrl = TaskDistributor.getSlaposHateoasUrl
      original_generateConfiguration = TaskDistributor.generateConfiguration
      original_isMasterTestnode = TaskDistributor.isMasterTestnode
      original_supply = SlapOSControler.supply
      original_request = SlapOSControler.request
      original_updateInstanceXML = RunnerClass._updateInstanceXML
      original_isHostingSubscriptionReady = SlapOSMasterCommunicator.isHostingSubscriptionReady
      original_isRegisteredHostingSubscription = SlapOSMasterCommunicator.isRegisteredHostingSubscription
      original_SlapOSMasterCommunicator__init__ = SlapOSMasterCommunicator.__init__
      TaskDistributor.getSlaposAccountKey = patch_getSlaposAccountKey
      TaskDistributor.getSlaposAccountCertificate = patch_getSlaposAccountCertificate
      TaskDistributor.getSlaposUrl = patch_getSlaposUrl
      TaskDistributor.getSlaposHateoasUrl = patch_getSlaposHateoasUrl
      TaskDistributor.generateConfiguration = patch_generateConfiguration
      TaskDistributor.isMasterTestnode = patch_isMasterTestnode
      SlapOSControler.supply = doNothing
      SlapOSControler.request = doNothing
      RunnerClass._updateInstanceXML = doNothing
      SlapOSMasterCommunicator.isHostingSubscriptionReady = patch_isHostingSubscriptionReady
      SlapOSMasterCommunicator.isRegisteredHostingSubscription = patch_isRegisteredHostingSubscription
      SlapOSMasterCommunicator.__init__ = doNothing
    original_startTestSuite = TaskDistributor.startTestSuite
    original_subscribeNode = TaskDistributor.subscribeNode
    original_getTestType = TaskDistributor.getTestType
    TaskDistributor.startTestSuite = patch_startTestSuite
    TaskDistributor.subscribeNode = doNothing
    TaskDistributor.getTestType = patch_getTestType
    original_createTestResult = TaskDistributionTool.createTestResult
    TaskDistributionTool.createTestResult = patch_createTestResult
    test_node = self.getTestNode()
    # Change UnitTestRunner class methods
    original_prepareSlapOS = RunnerClass._prepareSlapOS

    original_runTestSuite = RunnerClass.runTestSuite
 
    if my_test_type == "ScalabilityTest":
      RunnerClass.runTestSuite = patch_runTestSuite
    else:
      RunnerClass.runTestSuite = doNothing

    RunnerClass._prepareSlapOS = doNothing
    SlapOSControler.initializeSlapOSControler = doNothing
    test_node.run()
    self.assertEquals(counter, 3)
    checkTestSuite(test_node)
    time.sleep = original_sleep
    # Restore old class methods
    if my_test_type == "ScalabilityTest":
      TaskDistributor.getSlaposAccountKey = original_getSlaposAccountKey
      TaskDistributor.getSlaposAccountCertificate = original_getSlaposAccountCertificate
      TaskDistributor.getSlaposUrl = original_getSlaposUrl
      TaskDistributor.getSlaposHateoasUrl = original_getSlaposHateoasUrl
      TaskDistributor.generateConfiguration = original_generateConfiguration
      TaskDistributor.isMasterTestnode = original_isMasterTestnode
      SlapOSControler.supply =original_supply
      SlapOSControler.request = original_request
      SlapOSControler.updateInstanceXML = original_updateInstanceXML
      SlapOSMasterCommunicator.isHostingSubscriptionReady = original_isHostingSubscriptionReady
      SlapOSMasterCommunicator.isRegisteredHostingSubscription = original_isRegisteredHostingSubscription
      SlapOSMasterCommunicator.__init__ = original_SlapOSMasterCommunicator__init__
    TaskDistributor.startTestSuite = original_startTestSuite
    TaskDistributionTool.createTestResult = original_createTestResult
    TaskDistributionTool.subscribeNode = original_subscribeNode
    TaskDistributionTool.getTestType = original_getTestType
    RunnerClass._prepareSlapOS = original_prepareSlapOS
    RunnerClass.runTestSuite = original_runTestSuite

  def test_16_cleanupLogDirectory(self):
    # Make sure that we are able to cleanup old log folders
    test_node = self.getTestNode()
    def check(file_list):
      log_directory_dir = os.listdir(self.log_directory)
      self.assertTrue(set(file_list).issubset(
           set(log_directory_dir)),
           "%r not contained by %r" % (file_list, log_directory_dir))
    check([])
    os.mkdir(os.path.join(self.log_directory, 'ab-llzje'))
    a_file = open(os.path.join(self.log_directory, 'a_file'), 'w')
    a_file.close()
    check(set(['ab-llzje', 'a_file']))
    # default log file time is 15 days, so nothing is going to be deleted
    test_node._cleanupLog()
    check(set(['ab-llzje', 'a_file']))
    # then we set keep time to 0, folder will be deleted
    test_node.max_log_time = 0
    test_node._cleanupLog()
    check(set(['a_file']))

  def test_17_cleanupTempDirectory(self):
    # Make sure that we are able to cleanup old temp folders
    test_node = self.getTestNode()
    temp_directory = self.system_temp_folder
    def check(file_list):
      directory_dir = os.listdir(temp_directory)
      self.assertTrue(set(file_list).issubset(
           set(directory_dir)),
           "%r not contained by %r" % (file_list, directory_dir))
    check([])
    os.mkdir(os.path.join(temp_directory, 'buildoutA'))
    os.mkdir(os.path.join(temp_directory, 'something'))
    os.mkdir(os.path.join(temp_directory, 'tmpC'))
    check(set(['buildoutA', 'something', 'tmpC']))
    # default log file time is 15 days, so nothing is going to be deleted
    test_node._cleanupTemporaryFiles()
    check(set(['buildoutA', 'something', 'tmpC']))
    # then we set keep time to 0, folder will be deleted
    test_node.max_temp_time = 0
    test_node._cleanupTemporaryFiles()
    check(set(['something']))

  def test_18_resetSoftwareAfterManyBuildFailures(self, my_test_type='UnitTest'):
    """
    Check that after several building failures that the software is resetted
    """
    initial_initializeSlapOSControler = \
      SlapOSControler.initializeSlapOSControler
    initial_runSoftwareRelease = SlapOSControler.runSoftwareRelease
    test_node = self.getTestNode()
    RunnerClass = self.returnGoodClassRunner(my_test_type)
    runner = RunnerClass(test_node)    
    node_test_suite = test_node.getNodeTestSuite('foo')
    init_call_kw_list = []
    def initializeSlapOSControler(self, **kw):
      init_call_kw_list.append(kw)
    def runSoftwareRelease(self, *args, **kw):
      return {"status_code": 1}
    SlapOSControler.initializeSlapOSControler = initializeSlapOSControler
    SlapOSControler.runSoftwareRelease = runSoftwareRelease
    def callPrepareSlapOS():
      runner._prepareSlapOS(self.working_directory, node_test_suite,
         test_node.log, create_partition=0)
    def callRaisingPrepareSlapos():
      self.assertRaises(SubprocessError, callPrepareSlapOS)
  
    self.assertEquals(node_test_suite.retry_software_count, 0)
    for x in xrange(0,11):
      callRaisingPrepareSlapos()
    self.assertEquals(len(init_call_kw_list), 11)
    self.assertEquals(init_call_kw_list[-1]['reset_software'], False)
    self.assertEquals(node_test_suite.retry_software_count, 11)
    callRaisingPrepareSlapos()
    self.assertEquals(init_call_kw_list[-1]['reset_software'], True)
    self.assertEquals(node_test_suite.retry_software_count, 1)
    callRaisingPrepareSlapos()
    self.assertEquals(init_call_kw_list[-1]['reset_software'], False)
    self.assertEquals(node_test_suite.retry_software_count, 2)
    SlapOSControler.initializeSlapOSControler = \
      initial_initializeSlapOSControler
    SlapOSControler.runSoftwareRelease = initial_runSoftwareRelease

  def test_scalability_04_constructProfile(self, my_test_type='ScalabilityTest'):
    self.test_04_constructProfile(my_test_type)
  def test_scalability_05b_changeRepositoryBranch(self, my_test_type='ScalabilityTest'):
    self.test_05b_changeRepositoryBranch(my_test_type)
  def test_scalability_09_runTestSuite(self, my_test_type='ScalabilityTest'):
    # TODO : write own scalability test
    pass
  def test_scalability_10_prepareSlapOS(self, my_test_type='ScalabilityTest'):
    # TODO : write own scalability test
    # This case test may be dispensable on ScalabilityTest case
    # so..
    pass
  @unittest.skip("Not implemented")
  def test_scalability_as_master_11_run(self, my_test_type='ScalabilityTest'):
    self.test_11_run(my_test_type, grade='master')
  # TODO : add a test with master and a launchable testsuite -> patch a lot of methods
  @unittest.skip("Not implemented")
  def test_scalability_as_slave_11_run(self, my_test_type='ScalabilityTest'):
    self.test_11_run(my_test_type, grade='slave')
  @unittest.skip("Not implemented")
  def test_scalability_as_master_15_suite_log_directory(self, my_test_type='ScalabilityTest'):
    self.test_15_suite_log_directory(my_test_type, grade='master')
  @unittest.skip("Not implemented")
  def test_scalability_as_slave_15_suite_log_directory(self, my_test_type='ScalabilityTest'):
    self.test_15_suite_log_directory(my_test_type, grade='slave')
  def test_scalability_18_resetSoftwareAfterManyBuildFailures(self, my_test_type='ScalabilityTest'):
    # TODO : write own scalability test
    pass

  def test_zzzz_scalability_19_xxxx(self):
    # TODO : fill the dummy slapos answer
    # by patching isSoftwareReleaseReady method.
    def patch_createTestResult(self, revision, test_name_list, node_title,
            allow_restart=False, test_title=None, project_title=None):
      test_result_path = os.path.join(test_result_path_root, test_title)
      result =  TestResultProxy(self._proxy, self._retry_time,
                self._logger, test_result_path, node_title, revision)
      return result
    global startTestSuiteDone
    startTestSuiteDone = False
    def patch_startTestSuite(self,node_title,computer_guid='unknown'):
      config_list = []
      global startTestSuiteDone
      if not startTestSuiteDone:
        startTestSuiteDone = True
        config_list.append(test_self.getTestSuiteData(reference='foo')[0])
        config_list.append(test_self.getTestSuiteData(reference='bar')[0])
      else:
        raise StopIteration
      return json.dumps(config_list)
    def patch_isMasterTestnode(self, *args, **kw):
      return True
    def patch_generateConfiguration(self, *args, **kw):
      return json.dumps({"configuration_list": [{"ok":"ok"}], "involved_nodes_computer_guid"\
: ["COMP1", "COMP2", "COMP3"], "error_message": "No error.", "launcher_nodes_computer_guid": ["COMP1"], \
"launchable": True, "randomized_path" : "azertyuiop"})
    def doNothing(self, *args, **kw):
        pass
    def patch_getSlaposAccountKey(self, *args, **kw):
      return "key"
    def patch_getSlaposAccountCertificate(self, *args, **kw):
      return "Certificate"
    def patch_getSlaposUrl(self, *args, **kw):
      return "http://Foo"
    def patch_getSlaposHateoasUrl(self, *args, **kw):
      return "http://Foo"
    def patch_getTestType(self, *args, **kw):
      return "ScalabilityTest"
    def patch_isHostingSubscriptionReady(self, *args, **kw):
      return True
    def patch_isRegisteredHostingSubscription(self, *args, **kw):
      return True     
    def patch_runTestSuite(self, *args, **kw):
      return {'status_code':0}
    test_self = self
    test_result_path_root = os.path.join(test_self._temp_dir,'test/results')
    os.makedirs(test_result_path_root)
    self.generateTestRepositoryList()
    # Select the good runner to modify
    RunnerClass = self.returnGoodClassRunner('ScalabilityTest')
    # Patch methods
    original_sleep = time.sleep
    original_getSlaposAccountKey = TaskDistributor.getSlaposAccountKey
    original_getSlaposAccountCertificate = TaskDistributor.getSlaposAccountCertificate
    original_getSlaposUrl = TaskDistributor.getSlaposUrl
    original_getSlaposHateoasUrl = TaskDistributor.getSlaposHateoasUrl
    original_generateConfiguration = TaskDistributor.generateConfiguration
    original_isMasterTestnode = TaskDistributor.isMasterTestnode
    original_startTestSuite = TaskDistributor.startTestSuite
    original_subscribeNode = TaskDistributor.subscribeNode
    original_getTestType = TaskDistributor.getTestType
    original_createTestResult = TaskDistributionTool.createTestResult
    original_prepareSlapOS = RunnerClass._prepareSlapOS
    original_runTestSuite = RunnerClass.runTestSuite
    original_supply = SlapOSControler.supply
    original_request = SlapOSControler.request
    original_updateInstanceXML = SlapOSControler.updateInstanceXML
    original_isHostingSubscriptionReady = SlapOSMasterCommunicator.isHostingSubscriptionReady
    original_isRegisteredHostingSubscription = SlapOSMasterCommunicator.isRegisteredHostingSubscription
    original_SlapOSMasterCommunicator__init__ = SlapOSMasterCommunicator.__init__

    #
    time.sleep = doNothing
    TaskDistributor.getSlaposAccountKey = patch_getSlaposAccountKey
    TaskDistributor.getSlaposAccountCertificate = patch_getSlaposAccountCertificate
    TaskDistributor.getSlaposUrl = patch_getSlaposUrl
    TaskDistributor.getSlaposHateoasUrl = patch_getSlaposHateoasUrl
    TaskDistributor.generateConfiguration = patch_generateConfiguration
    TaskDistributor.isMasterTestnode = patch_isMasterTestnode
    TaskDistributor.startTestSuite = patch_startTestSuite
    TaskDistributor.subscribeNode = doNothing
    TaskDistributor.getTestType = patch_getTestType    
    TaskDistributionTool.createTestResult = patch_createTestResult
    RunnerClass._prepareSlapOS = doNothing
    RunnerClass.runTestSuite = patch_runTestSuite
    SlapOSControler.supply = doNothing
    SlapOSControler.request = doNothing
    SlapOSControler.updateInstanceXML = doNothing
    SlapOSMasterCommunicator.isHostingSubscriptionReady = patch_isHostingSubscriptionReady
    SlapOSMasterCommunicator.isRegisteredHostingSubscription = patch_isRegisteredHostingSubscription
    SlapOSMasterCommunicator.__init__ = doNothing
    # Run
    test_node = self.getTestNode()  
    test_node.run()
    # Restore methods
    TaskDistributor.getSlaposAccountKey = original_getSlaposAccountKey
    TaskDistributor.getSlaposAccountCertificate = original_getSlaposAccountCertificate
    TaskDistributor.getSlaposUrl = original_getSlaposUrl
    TaskDistributor.getSlaposHateoasUrl = original_getSlaposHateoasUrl
    TaskDistributor.generateConfiguration = original_generateConfiguration
    TaskDistributor.isMasterTestnode = original_isMasterTestnode
    TaskDistributor.startTestSuite = original_startTestSuite
    TaskDistributionTool.createTestResult = original_createTestResult
    TaskDistributionTool.subscribeNode = original_subscribeNode
    TaskDistributionTool.getTestType = original_getTestType
    RunnerClass._prepareSlapOS = original_prepareSlapOS
    RunnerClass.runTestSuite = original_runTestSuite
    SlapOSControler.supply = original_supply
    SlapOSControler.request = original_request
    SlapOSControler.updateInstanceXML = original_updateInstanceXML
    SlapOSMasterCommunicator.isHostingSubscriptionReady = original_isHostingSubscriptionReady
    SlapOSMasterCommunicator.isRegisteredHostingSubscription = original_isRegisteredHostingSubscription
    SlapOSMasterCommunicator.__init__ = original_SlapOSMasterCommunicator__init__
    time.sleep =original_sleep



#!/usr/bin/python
##############################################################################
#
# Copyright (c) 2009 Nexedi SA and Contributors. All Rights Reserved.
#                    Vincent Pelletier <vincent@nexedi.com>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import os
import sys
import imp
import gzip
import getopt
from time import time

PROFILING_ENABLED = False
if PROFILING_ENABLED:
  from tiny_profiler import profiler_decorator, profiler_report
else:
  def profiler_decorator(funct):
    return funct

  def profiler_report():
    pass

usage = """
Usage:
  parse_timing_log.py [--prefix <prefix>] --config <config> [--debug]
                      [--no-average] [--sum] [--load <file>] [--save <file>]
                      [--decimate <int>] [file_1 [file_2 [...]]]

  Either --prefix or --save must be given.

  --prefix <prefix>
    <prefix> is a string which is used to prefix result file names.
    If ommited, no CSV will be generated.

  --no-average
    Disable the generation of CSV files with average values.

  --sum
    Generate CSV files with time sum as values.
    They use the same names as average files, suffixed with "_sum.csv"
    Ignored if --prefix was not given.

  --load <file>
    Load internal data dict from given file before processing any given file.
    If it's given multiple time, the content of all those files will be merged.

  --save <file>
    Save interal data dict to given file after processing all given files.

  --config <config>
    <config> is a python script defining 2 values:
      - a method called "processLine"
      - a compiled regex called "LINE_PATTERN"
      - a date list sort key computation function called "date_key"
  
  --debug
    Display missed and skipped lines.

  --decimate <int>
    Instead of generating a line per measure, generate one line per <int>
    measures.
    Remain of the integer division of the number of measures per decimate value
    are all put in latest output line.

  file_1 ...
    Log files to process.
    Order in which files are given does not matter.
    Files can be gzip or plain text.

Output files:
  CSV, one file per distinct processLine return value, one line per log day,
  one column per measure.
  First line contains column titles.
  First column contains measure date (first recognisable date in current file).
  Each other cell contains:
    =<value sum>/<value count>
  Example:
    =434/125
  Which means an average of 3.472s over 125 values.
  Empty clls means that there are no values for that measure in current file.
  Strings are surrounded by double quotes (").
  Fields are sparated by colons (,).
"""

@profiler_decorator
def parseFile(filename, measure_dict):
  date = None
  line_number = 0
  match_count = 0
  skip_count = 0
  logfile = gzip.open(filename, 'r')
  try:
    line = logfile.readline()
  except IOError:
    logfile = open(filename, 'r')
    line = logfile.readline()
  begin = time()
  while line != '':
    line_number += 1
    if line_number % 5000 == 0:
      sys.stderr.write('%i\r' % (line_number, ))
      sys.stderr.flush()
    match_list = LINE_PATTERN.findall(line)
    if len(match_list) != 1:
      print >>sys.stderr, 'Unparseable line: %s:%i %r' % (filename, line_number, line)
    else:
      result, filter_id, date, duration = processLine(match_list[0], filename, line_number)
      # Possible result values & meaning:
      #  False: try next filter_method
      #  True: ignore & skip to next line
      #  (string): use & skip to next line
      if result is False:
        if debug:
          print >>sys.stderr, '? %s:%i %r' % (filename, line_number, match_list[0])
      elif result is True:
        if debug:
          print >>sys.stderr, '- %s:%i %r' % (filename, line_number, match_list[0])
        skip_count += 1
      else:
        measure_dict.setdefault(filter_id, {}).setdefault(result, {}).setdefault(date, []).append(int(duration))
        match_count += 1
    line = logfile.readline()
  print >>sys.stderr, '%i' % (line_number, )
  if line_number > 0:
    duration = time() - begin
    print >>sys.stderr, "Matched %i lines (%.2f%%), %i skipped (%.2f%%), %i unmatched (%.2f%%) in %.2fs (%i lines per second)." % \
      (match_count, (float(match_count) / line_number) * 100, skip_count, (float(skip_count) / line_number) * 100, (line_number - match_count - skip_count), (1 - (float(match_count + skip_count) / line_number)) * 100, duration, line_number / duration)

debug = False
outfile_prefix = None
configuration = None
do_average = True
do_sum = False
load_file_name_list = []
save_file_name = None
decimate_count = 1

try:
  opts, file_list = getopt.getopt(sys.argv[1:], '', ['debug', 'config=', 'prefix=', 'no-average', 'sum', 'load=', 'save=', 'decimate='])
except Exception, reason:
  print >>sys.stderr, reason
  print >>sys.stderr, usage
  sys.exit(1)

for name, value in opts:
  if name == '--debug':
    debug = True
  elif name == '--config':
    configuration = value
  elif name == '--prefix':
    outfile_prefix = value
  elif name == '--no-average':
    do_average = False
  elif name == '--sum':
    do_sum = True
  elif name == '--load':
    load_file_name_list.append(value)
  elif name == '--save':
    save_file_name = value
  elif name == '--decimate':
    decimate_count = int(value)

if configuration is None:
  raise ValueError, '--config is mandatory'

config_file = os.path.splitext(os.path.basename(configuration))[0]
config_path = [os.path.dirname(os.path.abspath(configuration))] + sys.path
file, path, description = imp.find_module(config_file, config_path)
module = imp.load_module(config_file, file, path, description)
file.close()

processLine = module.processLine
LINE_PATTERN = module.LINE_PATTERN
date_key = module.date_key

file_count = len(file_list)
file_number = 0

measure_dict = {}
if len(load_file_name_list):
  for load_file_name in load_file_name_list:
    load_file = open(load_file_name)
    temp_measure_dict = eval(load_file.read(), {})
    load_file.close()
    assert isinstance(measure_dict, dict)
    for filter_id, result_dict in temp_measure_dict.iteritems():
      for result, date_dict in result_dict.iteritems():
        for date, duration_list in date_dict.iteritems():
          measure_dict.setdefault(filter_id, {}).setdefault(result, {}).setdefault(date, []).extend(duration_list)
    print >>sys.stderr, 'Previous processing result restored from %r' % (load_file_name, )

for filename in file_list:
  file_number += 1
  print >>sys.stderr, 'Processing %s [%i/%i]...' % (filename, file_number, file_count)
  parseFile(filename, measure_dict)

if save_file_name is not None:
  save_file = open(save_file_name, 'w')
  save_file.write(repr(measure_dict))
  save_file.close()
  print >>sys.stderr, 'Processing result saved to %r' % (save_file_name, )

if outfile_prefix is not None:
  ## Generate a list of all measures and a 2-levels dictionnary with date as key and measure dictionnary as value
  measure_id_list = []
  append = measure_id_list.append
  sheet_dict = {}
  line_dict = {}
  for match_id, match_dict in measure_dict.iteritems():
    for result_id, result_dict in match_dict.iteritems():
      measure_id = (match_id, result_id)
      sheet_dict.setdefault(match_id, []).append((result_id, measure_id))
      append(measure_id)
      for date, measure_list in result_dict.iteritems():
        first_level_dict = line_dict.setdefault(date, {})
        assert measure_id not in first_level_dict
        first_level_dict[measure_id] = measure_list

  date_list = line_dict.keys()
  date_list.sort(key=date_key)

  def render_cell(value_list, format):
    if isinstance(value_list, (list, tuple)):
      return format % {'sum': sum(value_list), 'count': len(value_list)}
    else:
      return value_list

  def renderOutput(data_format, filename_suffix):
    for sheet_id, sheet_column_list in sheet_dict.iteritems():
      outfile_name = '%s_%s_%s.csv' % (outfile_prefix, sheet_id, filename_suffix)
      print >>sys.stderr, 'Writing to %r...' % (outfile_name, )
      outfile = open(outfile_name, 'w')
      print >>outfile, '"date",%s' % (','.join(['"%s"' % (x[0], ) for x in sheet_column_list]), )
      decimate_dict = {}
      decimate = 0
      for date in date_list:
        for key, value in line_dict[date].iteritems():
          decimate_dict.setdefault(key, []).extend(value)
        decimate += 1
        if decimate == decimate_count:
          print >>outfile, '"%s",%s' % (date, ','.join([render_cell(decimate_dict.get(x[1], ''), data_format) for x in sheet_column_list]))
          decimate_dict = {}
          decimate = 0
      if len(decimate_dict):
        print >>outfile, '"%s",%s' % (date, ','.join([render_cell(decimate_dict.get(x[1], ''), data_format) for x in sheet_column_list]))

  if do_average:
    renderOutput('=%(sum)i/%(count)i', 'avg')
  if do_sum:
    renderOutput('=%(sum)i', 'sum')

profiler_report()

